﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using Wsp.Types;

namespace Wsp.Algorithms
{
	abstract class MatchingBasedHeuristic
	{
		public MatchingBasedHeuristic()
		{
		}

		public int DistinctUsers(int[] pattern, int[] tasks)
		{
			int count = 0;
			var used = new Int64BasedSet();

			for (int i = 0; i < tasks.Length; i++)
			{
				int task = tasks[i];
				int user = pattern[task];
				if (used[user])
					continue;

				used[user] = true;
				count++;
			}

			return count;
		}

		public void DecodePattern(int[] encoding, int[] pattern)
		{
			int userCounter = 0;

			for (int i = 0; i < encoding.Length; i++)
			{
				Debug.Assert(encoding[i] >= 0);
				Debug.Assert(encoding[i] <= i);

				if (encoding[i] == i)
					pattern[i] = userCounter++;
				else
					pattern[i] = pattern[encoding[i]];
			}
		}

		public void RandomEncoding(int[] encoding, Random random)
		{
			for (int i = 0; i < encoding.Length; i++)
				encoding[i] = random.Next(i + 1);
		}

		protected int CalculateObjective(WspInstance instance, int[] pattern)
		{
			Debug.Assert(pattern.Length == instance.NumberOfSteps);

			int result = 0;

			for (int i = instance.NotEquals.Length - 1; i >= 0; i--)
			{
				int user1 = pattern[instance.NotEquals[i].Step1];
				int user2 = pattern[instance.NotEquals[i].Step2];
				if (user1 == user2)
					result++;
			}

			for (int i = instance.AtMost.Length - 1; i >= 0; i--)
			{
				int count = DistinctUsers(pattern, instance.AtMost[i].Steps);
				if (count > instance.AtMost[i].Users)
					result++;
			}

			for (int i = instance.AtLeast.Length - 1; i >= 0; i--)
			{
				int count = DistinctUsers(pattern, instance.AtLeast[i].Steps);
				if (count < instance.AtLeast[i].Users)
					result++;
			}

			return result;
		}

		public static int[] SolveMatchingProblem(bool[,] edges)
		{
			int n1 = edges.GetLength(0);
			int n2 = edges.GetLength(1);

			var match1 = (-1).FillArray(n1);
			var match2 = (-1).FillArray(n2);
			
			bool failed = false;
			for (int i = 0; i < n1; i++)
			{
				for (int j = 0; j < n2; j++)
				{
					if (match2[j] >= 0)
						continue;

					if (edges[i, j])
					{
						match1[i] = j;
						match2[j] = i;
						break;
					}
				}

				if (match1[i] < 0)
					failed = true;
			}

			if (failed)
			{
				for (int i = 0; i < n1; i++)
				{
					if (match1[i] < 0)
						if (!AugmentingPath(edges, match1, match2, i))
							return null;
				}
			}

			var used = new bool[n2];
			for (int i = 0; i < n1; i++)
			{
				int j = match1[i];

				Debug.Assert(edges[i, j]);
				Debug.Assert(!used[j]);

				used[j] = true;

				//Helper.DebugMessage("a{0} - b{1}", i, j);
			}

			return match1;
		}

		private static bool AugmentingPath(bool[,] edges, int[] match1, int[] match2, int i)
		{
			var visited1 = new bool[match1.Length];
			var visited2 = new bool[match2.Length];

			visited1[i] = true;
			var path = new List<int>();
			path.Add(i);

			if (DepthFirstSearch2(edges, path, visited1, visited2, match1, match2))
			{
				for (int k = 0; k < path.Count; k += 2)
					match1[path[k]] = path[k + 1];

				return true;
			}

			return false;
		}

		private static bool AugmentingPath(Func<int, int, bool> edge, int[] match1, int[] match2, int i)
		{
			var visited1 = new bool[match1.Length];
			var visited2 = new bool[match2.Length];

			visited1[i] = true;
			var path = new List<int>();
			path.Add(i);

			if (DepthFirstSearch2(edge, path, visited1, visited2, match1, match2))
			{
				for (int k = 0; k < path.Count; k += 2)
					match1[path[k]] = path[k + 1];

				return true;
			}

			return false;
		}

		public static int[] SolveMatchingProblem(int n1, int n2, Func<int, int, bool> edge)
		{
			var match1 = (-1).FillArray(n1);
			var match2 = (-1).FillArray(n2);
			
			bool failed = false;
			for (int i = 0; i < n1; i++)
			{
				bool atLeastOneEdgeExists = false;
				for (int j = 0; j < n2; j++)
				{
					if (!edge(i, j))
						continue;

					atLeastOneEdgeExists = true;
					
					if (match2[j] >= 0)
						continue;

					if (edge(i, j))
					{
						match1[i] = j;
						match2[j] = i;
						break;
					}
				}

				if (!atLeastOneEdgeExists)
					return null;

				if (match1[i] < 0)
					failed = true;
			}

			if (failed)
			{
				for (int i = 0; i < n1; i++)
				{
					if (match1[i] < 0)
						if (!AugmentingPath(edge, match1, match2, i))
							return null;
				}
			}

			var used = new bool[n2];
			for (int i = 0; i < n1; i++)
			{
				int j = match1[i];

				Debug.Assert(edge(i, j));
				Debug.Assert(!used[j]);

				used[j] = true;

				//Helper.DebugMessage("a{0} - b{1}", i, j);
			}

			return match1;
		}

		private static bool DepthFirstSearch2(bool[,] edges, List<int> path, bool[] visited1, bool[] visited2, int[] match1, int[] match2)
		{
			int i = path[path.Count - 1];

			for (int j = 0; j < visited2.Length; j++)
			{
				if (!edges[i, j])
					continue;

				if (visited2[j])
					continue;

				if (match1[i] == j)
					continue;

				path.Add(j);

				if (match2[j] < 0)
					return true;

				visited2[j] = true;

				if (DepthFirstSearch1(edges, path, visited1, visited2, match1, match2))
					return true;

				path.RemoveAt(path.Count - 1);
			}

			return false;
		}

		private static bool DepthFirstSearch2(Func<int, int, bool> edge, List<int> path, bool[] visited1, bool[] visited2, int[] match1, int[] match2)
		{
			int i = path[path.Count - 1];

			for (int j = 0; j < visited2.Length; j++)
			{
				if (!edge(i, j))
					continue;

				if (visited2[j])
					continue;

				if (match1[i] == j)
					continue;

				path.Add(j);

				if (match2[j] < 0)
					return true;

				visited2[j] = true;

				if (DepthFirstSearch1(edge, path, visited1, visited2, match1, match2))
					return true;

				path.RemoveAt(path.Count - 1);
			}

			return false;
		}

		private static bool DepthFirstSearch1(bool[,] edges, List<int> path, bool[] visited1, bool[] visited2, int[] match1, int[] match2)
		{
			int j = path[path.Count - 1];

			var i = match2[j];
			Debug.Assert(i >= 0);

			if (visited1[i])
				return false;

			path.Add(i);
			visited1[i] = true;

			if (DepthFirstSearch2(edges, path, visited1, visited2, match1, match2))
				return true;
			
			path.RemoveAt(path.Count - 1);
			
			return false;
		}

		private static bool DepthFirstSearch1(Func<int, int, bool> edge, List<int> path, bool[] visited1, bool[] visited2, int[] match1, int[] match2)
		{
			int j = path[path.Count - 1];

			var i = match2[j];
			Debug.Assert(i >= 0);

			if (visited1[i])
				return false;

			path.Add(i);
			visited1[i] = true;

			if (DepthFirstSearch2(edge, path, visited1, visited2, match1, match2))
				return true;
			
			path.RemoveAt(path.Count - 1);
			
			return false;
		}

		protected bool AssignUsers(WspInstance instance, int[] pattern, ref WspSolution solution)
		{
			int numberOfUsers = 0;
			for (int i = pattern.Length - 1; i >= 0; i--)
			{
				if (pattern[i] > numberOfUsers)
					numberOfUsers = pattern[i];
			}

			numberOfUsers++; // The number of users is the index of the last user plus one.

			var edges = new bool[numberOfUsers, instance.NumberOfUsers];

			for (int assignedUser = 0; assignedUser < numberOfUsers; assignedUser++)
			{
				for (int user = 0; user < instance.NumberOfUsers; user++)
					edges[assignedUser, user] = true;

				for (int task = 0; task < pattern.Length; task++)
					if (pattern[task] == assignedUser)
					{
						for (int user = 0; user < instance.NumberOfUsers; user++)
							if (!instance.AuthorisationLists[user][task])
								edges[assignedUser, user] = false;
					}
			}

			var match = SolveMatchingProblem(edges);
			if (match == null)
				return false;

			for (int task = 0; task < pattern.Length; task++)
				solution.AssignedUsers[task] = match[pattern[task]];

			Debug.Assert(solution.IsFeasible());

			return true;
		}

		protected int[] AssignUsersFunc(WspInstance instance, int[] pattern)
		{
			int numberOfUsers = 0;
			for (int i = pattern.Length - 1; i >= 0; i--)
			{
				if (pattern[i] > numberOfUsers)
					numberOfUsers = pattern[i];
			}

			numberOfUsers++; // The number of users is the index of the last user plus one.

			var tasks = new int[numberOfUsers][];
			for (int i = 0; i < numberOfUsers; i++)
			{
				var t = new List<int>();
				for (int task = 0; task < pattern.Length; task++)
				{
					if (pattern[task] == i)
						t.Add(task);
				}

				tasks[i] = t.ToArray();
			}

			var match = SolveMatchingProblem(numberOfUsers, instance.NumberOfUsers,
											(t, u) =>
											{
												for (int i = 0; i < tasks[t].Length; i++)
													if (!instance.AuthorisationLists[u][tasks[t][i]])
														return false;

												return true;
											});

			if (match == null)
				return null;

			var assignment = new int[instance.NumberOfSteps];
			for (int step = 0; step < pattern.Length; step++)
				assignment[step] = match[pattern[step]];

			return assignment;
		}
	}
}
