﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using Wsp.Types;

namespace Wsp.Algorithms
{
	class Pbt : MatchingBasedHeuristic
	{
		private readonly TimeInterval givenTime;

#if COUNTERS
		public static int NodesCounter;
	    public static long UsersIterations;
	    public static int FullScan;
        public static PerformanceTimeCounter UsersLookupWatch = new PerformanceTimeCounter();
        public static PerformanceTimeCounter MatchingWatch = new PerformanceTimeCounter();
	    public static int BfsCalls;
	    public static int BfsTotalDepth;
        public static int InnermostBfsCounter;
#endif

	    public Pbt()
			: this(TimeInterval.Infinity)
		{
			givenTime = TimeInterval.Infinity;
		}

	    public TimeInterval GivenTime
	    {
	        get { return givenTime; }
	    }

	    public Pbt(TimeInterval givenTime)
		{
			this.givenTime = givenTime;
		}

		public WspSolution Solve(WspInstance instance)
		{
			var context = new Context(instance, givenTime);
			if (!Solve(context, new Int64BasedSet()))
				return new WspSolution(instance, context.ElapsedTime, WspSolutionType.ProvenUnsat);

            context.SaveBranchingFactors();

			if (context.Assignment != null)
				return new WspSolution(instance, context.ElapsedTime, context.Assignment);

			return new WspSolution(instance, context.ElapsedTime);
		}

		private bool Solve(Context context, Int64BasedSet assigned)
		{
#if COUNTERS
		    NodesCounter++;
#endif
			var instance = context.Instance;
			var pattern = context.Pattern;

			// DEBUG

			Debug.Assert(context.BlockCountersCorrect(assigned));
			
			/*
						using (var writer = File.AppendText("PB.txt"))
						{
							for (int i = 0; i < instance.NumberOfSteps; i++)
							{
								if (assigned[i])
									writer.Write("{0}\t", pattern[i]);
								else
									writer.Write("-1\t");
							}

							writer.WriteLine();
						}
			*/
			// END OF DEBUG


			// All the steps assigned
			if (assigned == Int64BasedSet.All(instance.NumberOfSteps))
			{
				context.Assignment = new int[instance.NumberOfSteps];
				for (int i = 0; i < instance.NumberOfSteps; i++)
					context.Assignment[i] = context.MatchingGraph.GetAssignedUser(pattern[i]);

				return true;
			}

			// Time is over
			if (givenTime != TimeInterval.Infinity && assigned.NumberOfElements == 20 && context.TimeIsOver)
				return true;

			// How many blocks are in the pattern
			int numberOfBlocks = 0;
			for (int i = instance.NumberOfSteps - 1; i >= 0; i--)
				if (assigned[i] && pattern[i] >= numberOfBlocks)
					numberOfBlocks = pattern[i] + 1;

			int stepToAssign = context.SelectNextStep(assigned);


//#if DEBUG
			var assignedExclusive = assigned;
//#endif

			var blocksToTry = context.CalculateAllowedBlocks(assigned, numberOfBlocks, stepToAssign);

			assigned[stepToAssign] = true;

			for (int block = 0; !blocksToTry.IsEmpty; block++)
			{
				if (!blocksToTry[block])
					continue;

				blocksToTry[block] = false;

				pattern[stepToAssign] = block;

				var nodeStepsInclusive = new Int64BasedSet();
				for (int i = 0; i < context.Pattern.Length; i++)
					if (assigned[i] && context.Pattern[i] == block)
						nodeStepsInclusive[i] = true;

				var nodeStepsExclusive = nodeStepsInclusive;
				nodeStepsExclusive[stepToAssign] = false;

				for (int i = context.InvolvedAtMostIndices[stepToAssign].Length - 1; i >= 0; i--)
				{
					int index = context.InvolvedAtMostIndices[stepToAssign][i];
					if (!instance.AtMost[index].StepsMask.Intersects(nodeStepsExclusive))
						context.AssignedBlocksAtMost[index]++;
				}

				for (int i = context.InvolvedAtLeastIndices[stepToAssign].Length - 1; i >= 0; i--)
				{
					int index = context.InvolvedAtLeastIndices[stepToAssign][i];
					if (!instance.AtLeast[index].StepsMask.Intersects(nodeStepsExclusive))
						context.AssignedBlocksAtLeast[index]++;
				}

				Debug.Assert(context.MatchingGraph.IsFeasible);
				Debug.Assert(context.BlockCountersCorrect(assigned));

				bool broken = false;
				if (block == numberOfBlocks)
				{
				    //bool expected = context.OldMatchingGraph.AddNode(Int64BasedSet.OneElementSet(stepToAssign));

					if (!context.MatchingGraph.AddNode(Int64BasedSet.OneElementSet(stepToAssign)))
					{
                        //Debug.Assert(!expected);
						//context.MatchingGraph.RemoveLastNode(); // Why if adding a node was unsuccessful?
						broken = true;
					}
                    //Debug.Assert(expected);
                }
				else
				{
				    //bool expected = context.OldMatchingGraph.AddStepsToNode(block, nodeStepsInclusive);
					if (!context.MatchingGraph.AddStepsToNode(block, nodeStepsInclusive))
					{		
						//var hasToBeTrue = context.MatchingGraph.AddStepsToNode(block, nodeStepsExclusive);
						//Debug.Assert(hasToBeTrue);
						broken = true;
                        //Debug.Assert(!expected);
                        //context.OldMatchingGraph.AddStepsToNode(block, nodeStepsExclusive);
					}
                    //else
                    //    Debug.Assert(expected);

					Debug.Assert(context.MatchingGraph.IsFeasible);
				}

				Debug.Assert(context.MatchingGraph.IsFeasible);

				if (broken)
				{
					for (int i = context.InvolvedAtMostIndices[stepToAssign].Length - 1; i >= 0; i--)
					{
						int index = context.InvolvedAtMostIndices[stepToAssign][i];
						if (!instance.AtMost[index].StepsMask.Intersects(nodeStepsExclusive))
							context.AssignedBlocksAtMost[index]--;
					}

					for (int i = context.InvolvedAtLeastIndices[stepToAssign].Length - 1; i >= 0; i--)
					{
						int index = context.InvolvedAtLeastIndices[stepToAssign][i];
						if (!instance.AtLeast[index].StepsMask.Intersects(nodeStepsExclusive))
							context.AssignedBlocksAtLeast[index]--;
					}

					Debug.Assert(context.BlockCountersCorrect(assignedExclusive));
					continue;
				}

				if (Solve(context, assigned))
					return true;

			    if (block == numberOfBlocks)
			    {
			        context.MatchingGraph.RemoveLastNode();
			        //context.OldMatchingGraph.RemoveLastNode();
			    }
			    else
			    {
                    //context.OldMatchingGraph.RemoveStepsFromNode(block, nodeStepsExclusive);
                    context.MatchingGraph.RemoveStepsFromNode(block, nodeStepsExclusive);
			    }

				{
					for (int i = context.InvolvedAtMostIndices[stepToAssign].Length - 1; i >= 0; i--)
					{
						int index = context.InvolvedAtMostIndices[stepToAssign][i];
						if (!instance.AtMost[index].StepsMask.Intersects(nodeStepsExclusive))
							context.AssignedBlocksAtMost[index]--;
					}

					for (int i = context.InvolvedAtLeastIndices[stepToAssign].Length - 1; i >= 0; i--)
					{
						int index = context.InvolvedAtLeastIndices[stepToAssign][i];
						if (!instance.AtLeast[index].StepsMask.Intersects(nodeStepsExclusive))
							context.AssignedBlocksAtLeast[index]--;
					}
				}

				Debug.Assert(context.MatchingGraph.IsFeasible);
				Debug.Assert(context.BlockCountersCorrect(assignedExclusive));
			}

			return false;
		}


		class Context
		{
			public readonly WspInstance Instance;
			public readonly int[] Pattern;
			private readonly AccurateRealTimeCounter timeCounter;
			private readonly TimeInterval givenTime;
			public int[] Assignment;
			public readonly IMatchingGraph MatchingGraph;
			//public readonly MatchingGraph OldMatchingGraph;
			public readonly int[] AssignedBlocksAtMost;
			public readonly int[] AssignedBlocksAtLeast;

			public readonly int[][] InvolvedNotEqualsIndices;
			public readonly int[][] InvolvedAtMostIndices;
			public readonly int[][] InvolvedAtLeastIndices;

			private readonly int[] fixedStepWeights;

			private readonly int[] rankings;

			public Context(WspInstance instance, TimeInterval givenTime)
			{
#if SAVE_BRANCHING_FACTOR
                branchingFactor = new long[instance.NumberOfSteps,instance.NumberOfSteps];
#endif

				Instance = instance;
				this.givenTime = givenTime;
				timeCounter = AccurateRealTimeCounter.StartNew();
				Pattern = new int[instance.NumberOfSteps];
				MatchingGraph = instance.NumberOfUsers <= Helper.Sqr(instance.NumberOfSteps)
					? (IMatchingGraph)new MatchingGraphOn(instance)
					: new MatchingGraphOk(instance);
				//OldMatchingGraph = new MatchingGraph(instance);


				InvolvedNotEqualsIndices = new int[instance.NumberOfSteps][];
				InvolvedAtMostIndices = new int[instance.NumberOfSteps][];
				InvolvedAtLeastIndices = new int[instance.NumberOfSteps][];
				for (int step = 0; step < instance.NumberOfSteps; step++)
				{
					{ // not-equals
						var list = new List<int>();
						for (int i = 0; i < instance.NotEquals.Length; i++)
							if (instance.NotEquals[i].StepsMask[step])
								list.Add(i);

						InvolvedNotEqualsIndices[step] = list.ToArray();
					}

					{ // at-most
						var list = new List<int>();
						for (int i = 0; i < instance.AtMost.Length; i++)
							if (instance.AtMost[i].StepsMask[step])
								list.Add(i);

						InvolvedAtMostIndices[step] = list.ToArray();
					}

					{ // at-least
						var list = new List<int>();
						for (int i = 0; i < instance.AtLeast.Length; i++)
							if (instance.AtLeast[i].StepsMask[step])
								list.Add(i);

						InvolvedAtLeastIndices[step] = list.ToArray();
					}
				}

				AssignedBlocksAtMost = new int[instance.AtMost.Length];
				AssignedBlocksAtLeast = new int[instance.AtLeast.Length];
				
				fixedStepWeights = new int[instance.NumberOfSteps];
				for (int step = 0; step < instance.NumberOfSteps; step++)
				{
					fixedStepWeights[step] = InvolvedNotEqualsIndices[step].Length * 15;
											//+ InvolvedAtMostIndices[step].Length * parameters[1]
											//+ InvolvedAtLeastIndices[step].Length * parameters[2];
				}

				for (int i = 0; i < instance.NotEquals.Length; i++)
				{
					var notEqual = instance.NotEquals[i];
					for (int j = 0; j < instance.AtMost.Length; j++)
					{
						if (instance.AtMost[j].StepsMask.IsSuperset(notEqual.StepsMask))
						{
							fixedStepWeights[notEqual.Step1] += 20;
							fixedStepWeights[notEqual.Step2] += 20;
						}
					}
				}

				for (int i = 0; i < instance.AtLeast.Length; i++)
				{
					var atLeast = instance.AtLeast[i];
					for (int j = 0; j < instance.AtMost.Length; j++)
					{
						var intersection = instance.AtMost[j].StepsMask.Intersection(atLeast.StepsMask);
						if (intersection.NumberOfElements >= 3)
							foreach (var step in intersection)
								fixedStepWeights[step] += 10;
					}
				}
/*
				for (int step = 0; step < fixedStepWeights.Length; step++)
				{
					fixedStepWeights[step] = InvolvedNotEqualsIndices[step].Length
											* InvolvedAtMostIndices[step].Length/*
											* InvolvedAtLeastIndices[step].Length#1# / parameters[6];
				}
*/

				rankings = new int[Instance.AtMost.Length];
			}

			public bool TimeIsOver
			{
				get { return givenTime != TimeInterval.Zero && timeCounter.CurrentElapsed >= givenTime; }
			}

			public TimeInterval ElapsedTime
			{
				get { return timeCounter.CurrentElapsed; }
			}

			public Int64BasedSet CalculateAllowedBlocks(Int64BasedSet assigned, int numberOfBlocks, int step)
			{
				Int64BasedSet blocksToTry = Int64BasedSet.All(numberOfBlocks + 1);

				for (int i = InvolvedNotEqualsIndices[step].Length - 1; i >= 0; i--)
				{
					int index = InvolvedNotEqualsIndices[step][i];
					var step1 = Instance.NotEquals[index].Step1;
					if (assigned[step1])
						blocksToTry[Pattern[step1]] = false;
					else
					{
						var step2 = Instance.NotEquals[index].Step2;
						if (assigned[step2])
							blocksToTry[Pattern[step2]] = false;
					}
				}

				for (int i = InvolvedAtMostIndices[step].Length - 1; i >= 0; i--)
				{
					int index = InvolvedAtMostIndices[step][i];
					int already = AssignedBlocksAtMost[index];
					if (Instance.AtMost[index].Users == already)
					{
						var assignedUsers = new Int64BasedSet();
						for (int j = Instance.AtMost[index].Steps.Length - 1; j >= 0; j--)
						{
							var s = Instance.AtMost[index].Steps[j];
							if (assigned[s])
								assignedUsers[Pattern[s]] = true;
						}

						blocksToTry = blocksToTry.Intersection(assignedUsers);
					}
				}

				for (int i = InvolvedAtLeastIndices[step].Length - 1; i >= 0; i--)
				{
					int index = InvolvedAtLeastIndices[step][i];
					int already = AssignedBlocksAtLeast[index];
/*
					int zeros = 0;
					for (int j = Instance.AtLeast[index].Steps.Length - 1; j >= 0; j--)
						if (!assigned[Instance.AtLeast[index].Steps[j]])
							zeros++;
*/
					int zeros = Instance.AtLeast[index].StepsMask.Exclude(assigned).NumberOfElements;
					if (zeros == Instance.AtLeast[index].Users - already)
						for (int j = Instance.AtLeast[index].Steps.Length - 1; j >= 0; j--)
						{
							var s = Instance.AtLeast[index].Steps[j];
							if (assigned[s])
								blocksToTry[Pattern[s]] = false;
						}
				}

#if SAVE_BRANCHING_FACTOR
			    branchingFactor[assigned.NumberOfElements, usersToTry.NumberOfElements]++;
#endif

                return blocksToTry;
			}

#if SAVE_BRANCHING_FACTOR
            private readonly long[,] branchingFactor;
#endif

			public int SelectNextStep(Int64BasedSet assigned)
			{
				int stepToAssign = -1;
				int bestRating = int.MinValue;

				for (int i = Instance.AtMost.Length - 1; i >= 0; i--)
				{
					int already = AssignedBlocksAtMost[i];
					switch (Instance.AtMost[i].Users - already)
					{
						case 0:
							rankings[i] = 200;
							break;

						case 1:
							rankings[i] = 20;
							break;

						default:
							rankings[i] = 0;
							break;
					}
				}

				for (int step = 0; step < Instance.NumberOfSteps; step++)
				{
					if (assigned[step])
						continue;

					int ranking = fixedStepWeights[step];

					for (int i = InvolvedAtMostIndices[step].Length - 1; i >= 0; i--)
					{
						int index = InvolvedAtMostIndices[step][i];
						ranking += rankings[index];
					}

					if (ranking > bestRating)
					{
						bestRating = ranking;
						stepToAssign = step;
					}
				}

				Debug.Assert(stepToAssign >= 0);
				return stepToAssign;
			}

			public int MinDistinctUsers(int[] tasks, Int64BasedSet assigned)
			{
				int count = 0;
				var used = new Int64BasedSet();

				for (int i = 0; i < tasks.Length; i++)
				{
					int task = tasks[i];
					if (assigned[task])
					{
						int user = Pattern[task];
						if (used[user])
							continue;

						used[user] = true;
						count++;
					}
				}

				return count;
			}

			public int MaxDistinctUsers(int[] tasks, Int64BasedSet assigned)
			{
				int count = 0;
				var used = new Int64BasedSet();

				for (int i = 0; i < tasks.Length; i++)
				{
					int task = tasks[i];
					if (assigned[task])
					{
						int user = Pattern[task];
						if (used[user])
							continue;

						used[user] = true;
						count++;
					}
					else
						count++;
				}

				return count;
			}

			public bool BlockCountersCorrect(Int64BasedSet assigned)
			{
				for (int i = 0; i < AssignedBlocksAtMost.Length; i++)
				{
					var users = new Int64BasedSet();
					for (int j = 0; j < Instance.NumberOfSteps; j++)
						if (assigned[j] && Instance.AtMost[i].StepsMask[j])
							users[Pattern[j]] = true;

					if (users.NumberOfElements != AssignedBlocksAtMost[i])
						return false;
				}

				for (int i = 0; i < AssignedBlocksAtLeast.Length; i++)
				{
					var users = new Int64BasedSet();
					for (int j = 0; j < Instance.NumberOfSteps; j++)
						if (assigned[j] && Instance.AtLeast[i].StepsMask[j])
							users[Pattern[j]] = true;

					if (users.NumberOfElements != AssignedBlocksAtLeast[i])
						return false;
				}

				return true;
			}

		    public void SaveBranchingFactors()
		    {
#if SAVE_BRANCHING_FACTOR
		        using (var writer = File.CreateText(Helper.GetFullFilename("branching-factor.dat")))
		        {
		            for (int depth = 0; depth < Instance.NumberOfSteps; depth++)
		            {
		                for (int factor = 0; factor < Instance.NumberOfSteps; factor++)
		                    writer.Write("{0}\t", branchingFactor[depth, factor]);

		                writer.WriteLine();
		            }
		        }
#endif
		    }
		}

		private interface IMatchingGraph
		{
			bool IsFeasible { get; }
			bool AddStepsToNode(int nodeIndex, Int64BasedSet newSteps);
			void RemoveStepsFromNode(int nodeIndex, Int64BasedSet newSteps);
			bool AddNode(Int64BasedSet nodeSteps);
			void RemoveLastNode();
			int GetAssignedUser(int leftElement);
		}

		class MatchingGraphOn : IMatchingGraph
		{
			private int numOfLeftNodes;
			private readonly WspInstance instance;
			private readonly List<int>[] incidenceLists;
			private readonly int[] matchingFromLeft;
			private readonly int[] matchingFromRight;

			private readonly int[] bfsBuffer;
			private readonly int[] bfsParent;
			private readonly bool[] visited;

			private readonly Dictionary<Int64BasedSet, List<int>> incidenceDictionary = new Dictionary<Int64BasedSet, List<int>>();

			public MatchingGraphOn(WspInstance instance)
			{
				this.instance = instance;
				incidenceLists = new List<int>[instance.NumberOfSteps];
				for (int i = 0; i < incidenceLists.Length; i++)
					incidenceLists[i] = new List<int>(instance.NumberOfUsers);

				matchingFromLeft = (-1).FillArray(instance.NumberOfSteps);
				matchingFromRight = (-1).FillArray(instance.NumberOfUsers);

				bfsBuffer = new int[instance.NumberOfUsers];
				bfsParent = new int[instance.NumberOfUsers];
				visited = new bool[instance.NumberOfUsers];
			}

			public bool IsFeasible
			{
				get
				{
					for (int leftElement = 0; leftElement < numOfLeftNodes; leftElement++)
					{
						if (matchingFromLeft[leftElement] < 0)
							return false;

						if (matchingFromRight[matchingFromLeft[leftElement]] != leftElement)
							return false;
					}

					for (int i = 0; i < matchingFromRight.Length; i++)
					{
						if (matchingFromRight[i] >= 0)
						{
							if (matchingFromRight[i] >= numOfLeftNodes)
								return false;

							if (matchingFromLeft[matchingFromRight[i]] != i)
								return false;
						}
					}

					return true;
				}
			}

			public bool AddStepsToNode(int nodeIndex, Int64BasedSet newSteps)
			{
#if COUNTERS
			    MatchingWatch.Resume();
#endif

				//Debug.Assert(matchingFromLeft[nodeIndex] >= 0);
				var oldList = incidenceLists[nodeIndex];
				incidenceLists[nodeIndex] = ComputeIncidenceList(newSteps);

				// If no authorised users exist, exit
			    if (incidenceLists[nodeIndex].Count == 0)
			    {
					incidenceLists[nodeIndex] = oldList;

#if COUNTERS
			        MatchingWatch.Stop();
#endif

					return false;
			    }

				// Reset the match from right to left (the left from right is reset inside BFS())
				var oldUser = matchingFromLeft[nodeIndex];
				if (oldUser >= 0)
					matchingFromRight[oldUser] = -1;

			    var bfs = BFS(nodeIndex);
				if (!bfs)
				{
					matchingFromLeft[nodeIndex] = oldUser;
					matchingFromRight[oldUser] = nodeIndex;
					incidenceLists[nodeIndex] = oldList;
				}


#if COUNTERS
				MatchingWatch.Stop();
#endif

				return bfs;
			}

			public void RemoveStepsFromNode(int nodeIndex, Int64BasedSet newSteps)
			{
#if COUNTERS
                MatchingWatch.Resume();
#endif

				incidenceLists[nodeIndex] = ComputeIncidenceList(newSteps);

#if COUNTERS
                MatchingWatch.Stop();
#endif

				// Current matching is supposed to be valid
				//Debug.Assert(incidenceLists[nodeIndex].Contains(matchingFromLeft[nodeIndex]));
			}

			public bool AddNode(Int64BasedSet nodeSteps)
			{
#if COUNTERS
                MatchingWatch.Resume();
#endif

                incidenceLists[numOfLeftNodes] = ComputeIncidenceList(nodeSteps);

				// If no authorised users exist, exit
			    if (incidenceLists[numOfLeftNodes].Count == 0)
			    {
#if COUNTERS
					MatchingWatch.Stop();
#endif
                    return false;
			    }

				if (BFS(numOfLeftNodes))
				{
					numOfLeftNodes++;

#if COUNTERS
					MatchingWatch.Stop();
#endif

                    return true;
				}

#if COUNTERS
                MatchingWatch.Stop();
#endif

				return false; // No matching exists
			}

/*
			private class IncidenceTreeNode
			{
				private IncidenceTreeNode included;
				private IncidenceTreeNode excluded;
				private List<int> list;

				public List<int> GetList(Int64BasedSet steps, int step, WspInstance instance)
				{
					if (step == instance.NumberOfSteps)
						return list;

					if (steps[step])
					{
						if (included == null)
							included = new IncidenceTreeNode();

						return included.GetList(steps, step + 1, instance);
					}
				}
			}
*/

		    private List<int> ComputeIncidenceList(Int64BasedSet nodeSteps)
		    {
#if COUNTERS
		        UsersLookupWatch.Resume();
#endif

		        //Stat[nodeSteps.NumberOfElements]++;

		        List<int> list;
		        if (incidenceDictionary.TryGetValue(nodeSteps, out list))
		        {
#if COUNTERS
                    UsersLookupWatch.Stop();
#endif

		            return list;
		        }

		        var incidenceList = new List<int>();
		        for (int user = instance.NumberOfUsers - 1; user >= 0; user--)
		        {
#if COUNTERS
		            UsersIterations++;
#endif

		            if (instance.AuthorisationLists[user].IsSuperset(nodeSteps))
		            {
		                incidenceList.Add(user);
		                if (incidenceList.Count >= instance.NumberOfSteps)
		                    break;
		            }
		        }

#if COUNTERS
				if (incidenceList.Count < instance.NumberOfSteps)
		            FullScan++;
#endif

		        if (incidenceDictionary.Count > 16*1024)
		            incidenceDictionary.Clear();

		        incidenceDictionary.Add(nodeSteps, incidenceList);

#if COUNTERS
		        UsersLookupWatch.Stop();
#endif

		        return incidenceList;
		    }

		    private bool BFS(int nodeIndex)
			{
#if COUNTERS
                BfsCalls++;
#endif

// BFS for an augmenting path
				Array.Clear(visited, 0, visited.Length);
				int curIndex = 0;
				int bufferSize = 0;
				Enque(nodeIndex, ref bufferSize);

				matchingFromLeft[nodeIndex] = -1;
				while (curIndex < bufferSize)
				{
					int user = bfsBuffer[curIndex];
					if (matchingFromRight[user] < 0)
					{
						do
						{
							var curLeftElement = bfsParent[user];
							int prevUser = matchingFromLeft[curLeftElement];
							matchingFromRight[user] = curLeftElement;
							matchingFromLeft[curLeftElement] = user;
							user = prevUser;
#if COUNTERS
						    InnermostBfsCounter++;
#endif
						} while (user >= 0);

#if COUNTERS
					    BfsTotalDepth += curIndex + 1;
#endif

						return true;
					}

					int leftElement = matchingFromRight[user];
					Enque(leftElement, ref bufferSize);

					curIndex++;
				}

#if COUNTERS
                BfsTotalDepth += curIndex;
#endif

				return false;
			}

			private void Enque(int leftNode, ref int bufferSize)
			{
				var incidenceList = incidenceLists[leftNode];
				for (int i = 0; i < incidenceList.Count; i++)
				{
					var user = incidenceList[i];
					if (visited[user])
						continue;

					visited[user] = true;
					bfsBuffer[bufferSize] = user;
					bfsParent[user] = leftNode;
					bufferSize++;
				}
			}

			public void RemoveLastNode()
			{
				numOfLeftNodes--;
				int user = matchingFromLeft[numOfLeftNodes];
				//matchingFromLeft[numOfLeftNodes] = -1; // Do not need
				matchingFromRight[user] = -1;
			}

			public int GetAssignedUser(int leftElement)
			{
				return matchingFromLeft[leftElement];
			}
		}

		class MatchingGraphOk : IMatchingGraph
		{
			private int numOfLeftNodes;
			private readonly WspInstance instance;
			private readonly List<int>[] incidenceLists;
			private readonly int[] matchingFromLeft;
			private readonly int[] bfsBuffer;

			private readonly Dictionary<Int64BasedSet, List<int>> incidenceDictionary = new Dictionary<Int64BasedSet, List<int>>();
			private readonly Dictionary<int, UserNode> rightNodes = new Dictionary<int, UserNode>();



			struct UserNode
			{
				public readonly int EdgesCount;
                public readonly int PrevLeftNodeInPath;
                public readonly int MatchedLeftNode;

				public UserNode(int edgesCount)
                {
					this.EdgesCount = edgesCount;
				    MatchedLeftNode = -1;
				    PrevLeftNodeInPath = -1;
                }

			    private UserNode(int edgesCount, int prevLeftNodeInPath, int matchedLeftNode)
			    {
			        this.EdgesCount = edgesCount;
			        PrevLeftNodeInPath = prevLeftNodeInPath;
			        MatchedLeftNode = matchedLeftNode;
			    }

			    public UserNode AddEdge()
				{
					return new UserNode(EdgesCount + 1, PrevLeftNodeInPath, MatchedLeftNode);
				}

                public UserNode RemoveEdge()
				{
                    return new UserNode(EdgesCount - 1, PrevLeftNodeInPath, MatchedLeftNode);                    
				}

                public UserNode SetPrevLeftNodeInPath(int leftNode)
				{
                    return new UserNode(EdgesCount, leftNode, MatchedLeftNode);
				}

                public UserNode SetMatchedLeftNode(int leftNode)
				{
                    return new UserNode(EdgesCount, PrevLeftNodeInPath, leftNode);
				}
			}

			public MatchingGraphOk(WspInstance instance)
			{
				this.instance = instance;
				incidenceLists = new List<int>[instance.NumberOfSteps];
				for (int i = 0; i < incidenceLists.Length; i++)
					incidenceLists[i] = new List<int>(instance.NumberOfUsers);

				matchingFromLeft = (-1).FillArray(instance.NumberOfSteps);

				bfsBuffer = new int[instance.NumberOfSteps * instance.NumberOfSteps];
			}

			public bool IsFeasible
			{
				get
				{
					List<int> usersUsed = new List<int>();
					for (int leftNode = 0; leftNode < numOfLeftNodes; leftNode++)
					{
						if (matchingFromLeft[leftNode] < 0)
							return false;

                        if (!incidenceLists[leftNode].Contains(matchingFromLeft[leftNode]))
							return false;

						if (!rightNodes.ContainsKey(matchingFromLeft[leftNode]))
							return false;

						if (usersUsed.Contains(matchingFromLeft[leftNode]))
							return false;

						if (rightNodes[matchingFromLeft[leftNode]].MatchedLeftNode != leftNode)
							return false;

						usersUsed.Add(matchingFromLeft[leftNode]);
					}

				    foreach (var rightNode in rightNodes)
				    {
				        if (rightNode.Value.MatchedLeftNode >= 0)
				            if (matchingFromLeft[rightNode.Value.MatchedLeftNode] != rightNode.Key)
				                return false;
				    }

                    Dictionary<int, int> edgeCounters = new Dictionary<int, int>();
				    for (int i = 0; i < numOfLeftNodes; i++)
				    {
				        foreach (var user in incidenceLists[i])
				        {
				            if (edgeCounters.ContainsKey(user))
				                edgeCounters[user]++;
				            else
				                edgeCounters.Add(user, 1);
				        }
				    }

				    if (rightNodes.Count != edgeCounters.Count)
				        return false;

				    foreach (var counter in edgeCounters)
				    {
				        if (rightNodes[counter.Key].EdgesCount != counter.Value)
				            return false;
				    }

					return true;
				}
			}

			public bool AddStepsToNode(int nodeIndex, Int64BasedSet newSteps)
			{
                Debug.Assert(IsFeasible);
#if COUNTERS
			    MatchingWatch.Resume();
#endif

				//Debug.Assert(matchingFromLeft[nodeIndex] >= 0);

                int oldMatchedUser = matchingFromLeft[nodeIndex];

                rightNodes[oldMatchedUser] = rightNodes[oldMatchedUser].SetMatchedLeftNode(-1);

				var oldIncidenceList = incidenceLists[nodeIndex];
				RemoveEdges(oldIncidenceList, nodeIndex);

				var newIncidenceList = ComputeIncidenceList(newSteps);

				// If no authorised users exist, exit
			    if (newIncidenceList.Count == 0)
			    {
					AddEdges(oldIncidenceList);

                    // Restore old matching
                    rightNodes[oldMatchedUser] = rightNodes[oldMatchedUser].SetMatchedLeftNode(nodeIndex);

#if COUNTERS
					MatchingWatch.Stop();
#endif

                    Debug.Assert(IsFeasible);
                    return false;
			    }

                AddEdges(newIncidenceList);
				
				incidenceLists[nodeIndex] = newIncidenceList;

                var bfs = BFS(nodeIndex);
				if (!bfs)
				{
                    incidenceLists[nodeIndex] = oldIncidenceList;
				
                    RemoveEdges(newIncidenceList, nodeIndex);
					AddEdges(oldIncidenceList);

                    // Restore old matching 
				    matchingFromLeft[nodeIndex] = oldMatchedUser;
                    rightNodes[oldMatchedUser] = rightNodes[oldMatchedUser].SetMatchedLeftNode(nodeIndex);
				}

#if COUNTERS
				MatchingWatch.Stop();
#endif

                Debug.Assert(IsFeasible);
                return bfs;
			}

			private void AddEdges(List<int> incidenceList)
			{
				for (int i = incidenceList.Count - 1; i >= 0; i--)
				{
					int user = incidenceList[i];
					if (!rightNodes.ContainsKey(user))
						rightNodes.Add(user, new UserNode(1));
					else
                        rightNodes[user] = rightNodes[user].AddEdge();
				}
			}

			private void RemoveEdges(List<int> incidenceList, int ignoreLeftNode)
			{
				for (int i = incidenceList.Count - 1; i >= 0; i--)
				{
					int user = incidenceList[i];
				    var newNode = rightNodes[user].RemoveEdge();
				    if (newNode.EdgesCount == 0)
				    {
				        Debug.Assert(new Range(numOfLeftNodes).All(left => left == ignoreLeftNode || matchingFromLeft[left] != user));
				        rightNodes.Remove(user);
				    }
				    else
				        rightNodes[user] = newNode;
				}
			}

			public void RemoveStepsFromNode(int nodeIndex, Int64BasedSet newSteps)
			{
#if COUNTERS
                MatchingWatch.Resume();
#endif

				var oldIncidenceList = incidenceLists[nodeIndex];
                int oldMatchedUser = matchingFromLeft[nodeIndex];

                // Reset the match
                rightNodes[oldMatchedUser] = rightNodes[oldMatchedUser].SetMatchedLeftNode(-1);

                RemoveEdges(oldIncidenceList, nodeIndex);

				var newIncidenceList = ComputeIncidenceList(newSteps);
                AddEdges(newIncidenceList);

                // Restore matching from right
			    bool oldMatchInNewIncidenceList = false;
			    for (int i = newIncidenceList.Count - 1; i >= 0; i--)
			    {
                    if (newIncidenceList[i] == oldMatchedUser)
			        {
                        rightNodes[oldMatchedUser] = rightNodes[oldMatchedUser].SetMatchedLeftNode(nodeIndex);
			            oldMatchInNewIncidenceList = true;
			            break;
			        }
			    }

                if (!oldMatchInNewIncidenceList)
                {
                    // Can happen if newIncidenceList is bigger than k 
                    // and the first k elements don't include the previsously 
                    // matched user.
                    for (int i = 0; ; i++)
                    {
                        int user = newIncidenceList[i];
                        if (rightNodes[user].MatchedLeftNode < 0)
                        {
                            matchingFromLeft[nodeIndex] = user;
                            rightNodes[user] = rightNodes[user].SetMatchedLeftNode(nodeIndex);
                            break;
                        }
                    }
                }

                incidenceLists[nodeIndex] = newIncidenceList;

#if COUNTERS
				MatchingWatch.Stop();
#endif

                Debug.Assert(IsFeasible);

				// Current matching is supposed to be valid
				//Debug.Assert(incidenceLists[nodeIndex].Contains(matchingFromLeft[nodeIndex]));
			}

			public bool AddNode(Int64BasedSet nodeSteps)
			{
#if COUNTERS
                MatchingWatch.Resume();
#endif

				var incidenceList = ComputeIncidenceList(nodeSteps);

				// If no authorised users exist, exit
			    if (incidenceList.Count == 0)
			    {
#if COUNTERS
                    MatchingWatch.Stop();
#endif

                    Debug.Assert(IsFeasible);
                    return false;
			    }

				AddEdges(incidenceList);

				incidenceLists[numOfLeftNodes] = incidenceList;

				if (BFS(numOfLeftNodes))
				{
					numOfLeftNodes++;

#if COUNTERS
                    MatchingWatch.Stop();
#endif

                    Debug.Assert(IsFeasible);
                    return true;
				}

				RemoveEdges(incidenceList, -1);

#if COUNTERS
				MatchingWatch.Stop();
#endif

                Debug.Assert(IsFeasible);
                return false; // No matching exists
			}

/*
			private class IncidenceTreeNode
			{
				private IncidenceTreeNode included;
				private IncidenceTreeNode excluded;
				private List<int> list;

				public List<int> GetList(Int64BasedSet steps, int step, WspInstance instance)
				{
					if (step == instance.NumberOfSteps)
						return list;

					if (steps[step])
					{
						if (included == null)
							included = new IncidenceTreeNode();

						return included.GetList(steps, step + 1, instance);
					}
				}
			}
*/

		    private List<int> ComputeIncidenceList(Int64BasedSet nodeSteps)
		    {
#if COUNTERS
		        UsersLookupWatch.Resume();
#endif

		        //Stat[nodeSteps.NumberOfElements]++;

		        List<int> list;
		        if (incidenceDictionary.TryGetValue(nodeSteps, out list))
		        {
#if COUNTERS
                    UsersLookupWatch.Stop();
#endif

		            return list;
		        }

		        var incidenceList = new List<int>();
		        for (int user = instance.NumberOfUsers - 1; user >= 0; user--)
		        {
#if COUNTERS
		            UsersIterations++;
#endif

		            if (instance.AuthorisationLists[user].IsSuperset(nodeSteps))
		            {
		                incidenceList.Add(user);
		                if (incidenceList.Count >= instance.NumberOfSteps)
		                    break;
		            }
		        }

#if COUNTERS
		        if (incidenceList.Count < instance.NumberOfSteps)
		            FullScan++;
#endif

		        if (incidenceDictionary.Count > 16*1024)
		            incidenceDictionary.Clear();

		        incidenceDictionary.Add(nodeSteps, incidenceList);

#if COUNTERS
		        UsersLookupWatch.Stop();
#endif

		        return incidenceList;
		    }

		    private bool BFS(int nodeIndex)
			{
#if COUNTERS
                BfsCalls++;
#endif

// BFS for an augmenting path
				int curIndex = 0;
				int bufferSize = 0;
				Enque(nodeIndex, ref bufferSize);

				matchingFromLeft[nodeIndex] = -1;
				while (curIndex < bufferSize)
				{
					int user = bfsBuffer[curIndex];
					if (rightNodes[user].MatchedLeftNode < 0)
					{
						do
						{
							var curLeftNode = rightNodes[user].PrevLeftNodeInPath;
							int prevUser = matchingFromLeft[curLeftNode];
							rightNodes[user] = rightNodes[user].SetMatchedLeftNode(curLeftNode);
							matchingFromLeft[curLeftNode] = user;
							user = prevUser;

#if COUNTERS
							InnermostBfsCounter++;
#endif
						} while (user >= 0);

#if COUNTERS
					    BfsTotalDepth += curIndex + 1;
#endif

                        // Reset prevLeftNode links in modified nodes
					    while (--bufferSize >= 0)
					    {
					        var u = bfsBuffer[bufferSize];
					        rightNodes[u] = rightNodes[u].SetPrevLeftNodeInPath(-1);
					    }

						return true;
					}

					int leftNode = rightNodes[user].MatchedLeftNode;
					Enque(leftNode, ref bufferSize);

					curIndex++;
				}

#if COUNTERS
                BfsTotalDepth += curIndex;
#endif

                // Reset prevLeftNode links in modified nodes
                while (--bufferSize >= 0)
                {
                    var u = bfsBuffer[bufferSize];
                    rightNodes[u] = rightNodes[u].SetPrevLeftNodeInPath(-1);
                }

				return false;
			}

			private void Enque(int leftNode, ref int bufferSize)
			{
				var incidenceList = incidenceLists[leftNode];
				for (int i = 0; i < incidenceList.Count; i++)
				{
					var user = incidenceList[i];
					if (rightNodes[user].PrevLeftNodeInPath >= 0)
						continue;

					bfsBuffer[bufferSize] = user;
					rightNodes[user] = rightNodes[user].SetPrevLeftNodeInPath(leftNode);
					bufferSize++;
				}
			}

			public void RemoveLastNode()
			{
				numOfLeftNodes--;

			    var user = matchingFromLeft[numOfLeftNodes];
			    rightNodes[user] = rightNodes[user].SetMatchedLeftNode(-1);

				var incidenceList = incidenceLists[numOfLeftNodes];
                RemoveEdges(incidenceList, -1);

                Debug.Assert(IsFeasible);
            }

			public int GetAssignedUser(int leftElement)
			{
				return matchingFromLeft[leftElement];
			}
		}

	}
}
