﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading;
using Wsp.Types;

namespace Wsp.OtherSolvers
{
	class PbPbSolver
	{
		private readonly TimeInterval timeLimit;
	    private readonly bool cliqueConstraints;
	    private readonly bool cuttingPlanes;

        public PbPbSolver(bool cliqueConstraints = true)
            : this(TimeInterval.Infinity, cliqueConstraints)
		{
		}

		public PbPbSolver(TimeInterval timeLimit, bool cliqueConstraints = true, bool cuttingPlanes = false)
		{
			this.timeLimit = timeLimit;
			this.cliqueConstraints = cliqueConstraints;
			this.cuttingPlanes = cuttingPlanes;
		}

		private static int counter;

	    public bool CliqueConstraints
	    {
	        get { return cliqueConstraints; }
	    }

	    public WspSolution Solve(WspInstance instance)
		{
			string filename =
				Helper.GetFullFilename(
					Helper.GetSubdirectory(Helper.ProgramDirectory, "TempPbPb"),
					"temp" + Interlocked.Increment(ref counter),
					".obp");

			try
			{
				Dictionary<int, Tuple<int, int>> mDirectory;
				var xDirectory = WriteObp(instance, filename, out mDirectory);
				var solution = RunSolver(instance, filename, xDirectory, mDirectory);

				return solution;
			}
			finally
			{
				File.Delete(filename);
			}
		}

	    private WspSolution RunSolver(WspInstance instance, string filename, Dictionary<int, Tuple<int, int>> xDirectory,
	        Dictionary<int, Tuple<int, int>> mDirectory)
	    {
	        var process = new Process
	        {
	            StartInfo = new ProcessStartInfo("java")
	            {
	                Arguments =
	                    string.Format("-jar sat4j-pb.jar {2} {0} {1}",
	                        timeLimit == TimeInterval.Infinity ? "" : ((int) timeLimit.Seconds).ToString(),
	                        filename,
	                        cuttingPlanes ? "CuttingPlanes" : "Default"),
	                RedirectStandardOutput = true,
	                RedirectStandardError = true,
	                UseShellExecute = false,

	            }
	        };

	        var timeCounter = AccurateRealTimeCounter.StartNew();
	        process.Start();
	        while (true)
	        {
	            var output = process.StandardOutput.ReadLine();
	            if (output == null)
	                break;

	            if (output.StartsWith("s"))
	            {
	                timeCounter.Stop();
	                switch (output.Substring(2).ToUpper())
	                {
	                    case "SATISFIABLE":
	                        break;

	                    case "UNSATISFIABLE":
	                        return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.ProvenUnsat);

	                    case "UNKNOWN":
	                        return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Unknown);

	                    default:
	                        throw new Exception(string.Format("Unknown SAT4J result: '{0}'.", output));
	                }
	            }
	            else if (output.StartsWith("v"))
	            {
	                bool[,] M = new bool[instance.NumberOfSteps, instance.NumberOfSteps];
	                var solution = new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Sat);
	                Helper.FillExistingArray(solution.AssignedUsers, -1);
	                var strings = output.Substring(2).Split(' ');
	                foreach (var str in strings)
	                {
	                    if (string.IsNullOrWhiteSpace(str))
	                        continue;

	                    bool minus = false;
	                    string varIndex;
	                    if (str.StartsWith("-"))
	                    {
	                        minus = true;
	                        varIndex = str.Substring(2);
	                    }
	                    else
	                        varIndex = str.Substring(1);

	                    int index = int.Parse(varIndex);
	                    if (xDirectory.ContainsKey(index))
	                    {
	                        int user = xDirectory[index].Item1;
	                        int step = xDirectory[index].Item2;
	                        Debug.Assert(instance.AuthorisationLists[user][step]);

	                        if (!minus)
	                            solution.AssignedUsers[step] = user;
	                    }

	                    if (mDirectory.ContainsKey(index) && !minus)
	                    {
	                        int s1 = mDirectory[index].Item1;
	                        int s2 = mDirectory[index].Item2;
	                        M[s1, s2] = M[s2, s1] = true;
	                    }
	                }

	                Debug.Assert(solution.IsFeasible());

	                return solution;
	            }
	            //else
	            //Console.WriteLine(output);
	        }

	        process.WaitForExit();

	        timeCounter.Stop();

	        return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Unknown);
	    }

	    struct SatConstraint
		{
			private readonly int[] leftVars;
			private readonly int[] rightVars;
			private readonly int rightConst;
			private readonly string relation;

			public SatConstraint(int[] leftVars, int[] rightVars, int rightConst, string relation)
			{
				this.leftVars = leftVars;
				this.rightVars = rightVars;
				this.rightConst = rightConst;
				this.relation = relation;
			}

			public void WriteConstraint(StreamWriter writer)
			{
				writer.WriteLine(
					"{0}{1} {2} {3};",
					Expression(leftVars, '+'),
					Expression(rightVars, '-'),
					relation,
					rightConst);
			}

			private static string Expression(IEnumerable<int> vars, char sign)
			{
				if (vars == null)
					return "";

				return " " + vars.ListToString(v => sign + "1 x" + v, " ");
			}
		}

	    public Dictionary<int, Tuple<int, int>> WriteObp(WspInstance instance, string filename, out Dictionary<int, Tuple<int, int>> mDirectory)
		{
			var xDirectory = new Dictionary<int, Tuple<int, int>>();
			mDirectory = new Dictionary<int, Tuple<int, int>>();

		    using (var writer = File.CreateText(filename))
		    {
		        int numberOfVariables = 0;
		        var constraints = new List<SatConstraint>();

		        var x = new int[instance.NumberOfUsers, instance.NumberOfSteps];
		        for (int user = 0; user < instance.NumberOfUsers; user++)
		            for (int step = 0; step < instance.NumberOfSteps; step++)
		            {
		                if (instance.AuthorisationLists[user][step])
		                {
		                    x[user, step] = ++numberOfVariables;
		                    xDirectory[numberOfVariables] = Tuple.Create(user, step);
		                }
		                else
		                    x[user, step] = -1;
		            }

		        // Authorisation (exactly 1 authorised user is assigned to each step)
		        for (int step = 0; step < instance.NumberOfSteps; step++)
		        {
		            var leftVars = new List<int>();
		            for (int user = 0; user < instance.NumberOfUsers; user++)
		                if (instance.AuthorisationLists[user][step])
		                    leftVars.Add(x[user, step]);

		            constraints.Add(new SatConstraint(leftVars.ToArray(), null, 1, "="));
		        }

		        // M variables
		        var M = new int[instance.NumberOfSteps, instance.NumberOfSteps];
		        for (int i = 0; i < instance.NumberOfSteps; i++)
		            for (int j = i + 1; j < instance.NumberOfSteps; j++)
		            {
		                M[i, j] = ++numberOfVariables;
		                mDirectory.Add(numberOfVariables, Tuple.Create(i, j));
		            }

                // Clique constraints (if M13 and M23 then M12, and if )
		        if (cliqueConstraints)
		            for (int step1 = 0; step1 < instance.NumberOfSteps; step1++)
		                for (int step2 = step1 + 1; step2 < instance.NumberOfSteps; step2++)
		                    for (int step3 = step2 + 1; step3 < instance.NumberOfSteps; step3++)
		                    {
		                        constraints.Add(new SatConstraint(
		                            new[] {M[step1, step2]},
		                            new[] {M[step1, step3], M[step2, step3]},
		                            -1,
		                            ">="));
		                        constraints.Add(new SatConstraint(
		                            new[] {M[step1, step2], M[step2, step3]},
		                            new[] {M[step1, step3]},
		                            1,
		                            "<="));
								constraints.Add(new SatConstraint(
									new[] { M[step1, step2], M[step1, step3] },
									new[] { M[step2, step3] },
									1,
									"<="));
							}

		        // Link between M and x
				for (int step1 = 0; step1 < instance.NumberOfSteps; step1++)
				{
					for (int step2 = step1 + 1; step2 < instance.NumberOfSteps; step2++)
					{
						int m = M[step1, step2];

						for (int user = 0; user < instance.NumberOfUsers; user++)
						{
							int x1 = x[user, step1];
							int x2 = x[user, step2];
							if (x1 >= 0)
							{
								if (x2 < 0)
									constraints.Add(new SatConstraint(new[] { x1, m }, null, 1, "<="));
								else
								{
									constraints.Add(new SatConstraint(new[] { x1, m }, new[] { x2 }, 1, "<="));
									constraints.Add(new SatConstraint(new[] { x2, m }, new[] { x1 }, 1, "<="));
									constraints.Add(new SatConstraint(new[] { x1, x2 }, new[] { m }, 1, "<="));
								}
							}
							else if (x2 >= 0)
								constraints.Add(new SatConstraint(new[] { x2, m }, null, 1, "<="));
						}
					}
				}

				// not-equals
				for (int c = 0; c < instance.NotEquals.Length; c++)
				{
					constraints.Add(
						new SatConstraint(
							new[] { M[instance.NotEquals[c].Step1, instance.NotEquals[c].Step2] },
							null,
							0,
							"="));
				}

				// at-most
				for (int c = 0; c < instance.AtMost.Length; c++)
				{
					if (instance.AtMost[c].Steps.Length != 5 || instance.AtMost[c].Users != 3)
						throw new Exception("Unsupported at-most constraint.");

					var leftVars = new List<int>();
					for (int i = 0; i < instance.AtMost[c].Steps.Length; i++)
						for (int j = i + 1; j < instance.AtMost[c].Steps.Length; j++)
							leftVars.Add(M[instance.AtMost[c].Steps[i], instance.AtMost[c].Steps[j]]);

					constraints.Add(new SatConstraint(leftVars.ToArray(), null, 2, ">="));
				}

				// at-least
				for (int c = 0; c < instance.AtLeast.Length; c++)
				{
					if (instance.AtLeast[c].Steps.Length != 5 || instance.AtLeast[c].Users != 3)
						throw new Exception("Unsupported at-least constraint.");

					var leftVars = new List<int>();
					for (int i = 0; i < instance.AtLeast[c].Steps.Length; i++)
						for (int j = i + 1; j < instance.AtLeast[c].Steps.Length; j++)
							leftVars.Add(M[instance.AtLeast[c].Steps[i], instance.AtLeast[c].Steps[j]]);

					constraints.Add(new SatConstraint(leftVars.ToArray(), null, 3, "<="));
				}

				writer.WriteLine("* #variable= {0} #constraint= {1}", numberOfVariables, constraints.Count);
				foreach (var constraint in constraints)
					constraint.WriteConstraint(writer);
			}

			return xDirectory;
		}
	}
}

