﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading;
using Wsp.Types;

namespace Wsp.OtherSolvers
{
	class UdPbSolver
	{
		private readonly TimeInterval timeLimit;

		public UdPbSolver() : this(TimeInterval.Infinity)
		{
		}

		public UdPbSolver(TimeInterval timeLimit)
		{
			this.timeLimit = timeLimit;
		}

		private static int counter;

		public WspSolution Solve(WspInstance instance)
		{
			string filename =
				Helper.GetFullFilename(
					Helper.GetSubdirectory(Helper.ProgramDirectory, "TempUdPb"),
					"temp" + Interlocked.Increment(ref counter),
					".obp");

			try
			{
				var vars = WriteObp(instance, filename);
				var solution = RunSolver(instance, filename, vars);

				return solution;
			}
			finally
			{
				File.Delete(filename);
			}
		}

		private WspSolution RunSolver(WspInstance instance, string filename, Dictionary<int, Tuple<int, int>> vars)
		{
			var process = new Process
			{
				StartInfo = new ProcessStartInfo("java")
				{
					Arguments = string.Format("-jar sat4j-pb.jar Default {0} {1}", timeLimit == TimeInterval.Infinity ? "" : ((int)timeLimit.Seconds).ToString(), filename),
					RedirectStandardOutput = true,
					RedirectStandardError = true,
					UseShellExecute = false,
				}
			};

			var timeCounter = AccurateRealTimeCounter.StartNew();
			process.Start();
			while (true)
			{
				var output = process.StandardOutput.ReadLine();
				if (output == null)
					break;

				if (output.StartsWith("s"))
				{
					timeCounter.Stop();
					switch (output.Substring(2).ToUpper())
					{
						case "SATISFIABLE":
							break;

						case "UNSATISFIABLE":
							return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.ProvenUnsat);

						case "UNKNOWN":
							return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Unknown);

						default:
							throw new Exception(string.Format("Unknown SAT4J result: '{0}'.", output));
					}
				}
				else if (output.StartsWith("v"))
				{
					var solution = new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Sat);
					Helper.FillExistingArray(solution.AssignedUsers, -1);
					var strings = output.Substring(2).Split(' ');
					foreach (var str in strings)
					{
						if (string.IsNullOrWhiteSpace(str))
							continue;

						bool minus = false;
						string varIndex;
						if (str.StartsWith("-"))
						{
							minus = true;
							varIndex = str.Substring(2);
						}
						else
							varIndex = str.Substring(1);

						int index = int.Parse(varIndex);
						if (vars.ContainsKey(index))
						{
							int user = vars[index].Item1;
							int step = vars[index].Item2;
							Debug.Assert(instance.AuthorisationLists[user][step]);

							if (!minus)
								solution.AssignedUsers[step] = user;
						}
					}

					Debug.Assert(solution.IsFeasible());

					return solution;
				}
				//else
					//Console.WriteLine(output);
			}

			process.WaitForExit();

			timeCounter.Stop();

			return new WspSolution(instance, timeCounter.Elapsed, WspSolutionType.Unknown);
		}

		struct SatConstraint
		{
			private readonly int[] leftVars;
			private readonly int[] rightVars;
			private readonly int rightConst;
			private readonly string relation;

			public SatConstraint(int[] leftVars, int[] rightVars, int rightConst, string relation)
			{
				this.leftVars = leftVars;
				this.rightVars = rightVars;
				this.rightConst = rightConst;
				this.relation = relation;
			}

			public void WriteConstraint(StreamWriter writer)
			{
				writer.WriteLine(
					"{0}{1} {2} {3};",
					Expression(leftVars, '+'),
					Expression(rightVars, '-'),
					relation,
					rightConst);
			}

			private static string Expression(IEnumerable<int> vars, char sign)
			{
				if (vars == null)
					return "";

				return " " + vars.ListToString(v => sign + "1 x" + v, " ");
			}
		}

	    public Dictionary<int, Tuple<int, int>> WriteObp(WspInstance instance, string filename)
		{
			var vars = new Dictionary<int, Tuple<int, int>>();

			using (var writer = File.CreateText(filename))
			{
				int numberOfVariables = 0;
				var constraints = new List<SatConstraint>();

				var x = new int[instance.NumberOfUsers, instance.NumberOfSteps];
				for (int user = 0; user < instance.NumberOfUsers; user++)
					for (int step = 0; step < instance.NumberOfSteps; step++)
					{
						if (instance.AuthorisationLists[user][step])
						{
							x[user, step] = ++numberOfVariables;
							vars[numberOfVariables] = Tuple.Create(user, step);
						}
						else
							x[user, step] = -1;
					}


				for (int step = 0; step < instance.NumberOfSteps; step++)
				{
					var leftVars = new List<int>();
					for (int user = 0; user < instance.NumberOfUsers; user++)
						if (instance.AuthorisationLists[user][step])
							leftVars.Add(x[user, step]);

					constraints.Add(new SatConstraint(leftVars.ToArray(), null, 1, "="));
				}

				// not-equals
				for (int c = 0; c < instance.NotEquals.Length; c++)
				{
					for (int user = 0; user < instance.NumberOfUsers; user++)
						if (instance.AuthorisationLists[user].IsSuperset(instance.NotEquals[c].StepsMask))
							constraints.Add(
								new SatConstraint(
									new[] { x[user, instance.NotEquals[c].Step1], x[user, instance.NotEquals[c].Step2] },
									null,
									1,
									"<="));
				}

				// at-most
				var y = new int[instance.NumberOfUsers, instance.AtMost.Length];
				for (int c = 0; c < instance.AtMost.Length; c++)
				{
					var leftVars = new int[instance.NumberOfUsers];
					for (int user = 0; user < instance.NumberOfUsers; user++)
					{
						y[user, c] = ++numberOfVariables;
						leftVars[user] = y[user, c];

						foreach (var step in instance.AtMost[c].Steps)
							if (instance.AuthorisationLists[user][step])
								constraints.Add(new SatConstraint(new[] { x[user, step] }, new[] { y[user, c] }, 0, "<="));
					}

					constraints.Add(new SatConstraint(leftVars, null, instance.AtMost[c].Users, "<="));
				}

				// at-least
				var z = new int[instance.NumberOfUsers, instance.AtLeast.Length];
				for (int c = 0; c < instance.AtLeast.Length; c++)
				{
					var leftVars = new int[instance.NumberOfUsers];
					for (int user = 0; user < instance.NumberOfUsers; user++)
					{
						z[user, c] = ++numberOfVariables;
						leftVars[user] = z[user, c];

						var rightVars = new List<int>();
						foreach (var step in instance.AtLeast[c].Steps)
							if (instance.AuthorisationLists[user][step])
								rightVars.Add(x[user, step]);

						constraints.Add(new SatConstraint(new[] { z[user, c] }, rightVars.ToArray(), 0, "<="));
					}

					constraints.Add(new SatConstraint(leftVars, null, instance.AtLeast[c].Users, ">="));
				}

				writer.WriteLine("* #variable= {0} #constraint= {1}", numberOfVariables, constraints.Count);
				foreach (var constraint in constraints)
					constraint.WriteConstraint(writer);
			}

			return vars;
		}
	}
}

