﻿using System;
using System.Collections.Generic;
using Wsp.Algorithms;
using Wsp.OtherSolvers;
using Wsp.Types;

namespace Wsp
{
    internal class Program
    {
        private static void Main(string[] args)
        {
            var prevColour = Console.ForegroundColor;
            Console.ForegroundColor = ConsoleColor.Yellow;
            Console.WriteLine(
                "This is a suite of tools for generation of workflow satisfiability instances, converting them into Pseudo-Boolean problems and solving using Pattern Backtracking algorithm.");
            Console.WriteLine("Version 1.0.");
            Console.WriteLine(
                "For details please refer to D. Karapetyan, A. Parkes, G. Gutin and A. Gagarin, Pattern-Based Approach to the Workflow Satisfiability Problem with User-Independent Constraints");
	        Console.WriteLine();
            Console.WriteLine("This tool is free for non-commercial use only.");
	        Console.WriteLine();
            Console.WriteLine("Contact details: daniel.karapetyan@gmail.com");
            Console.ForegroundColor = prevColour;
            Console.WriteLine();

            if (args.Length == 0)
            {
                HelpMessage();
                return;
            }

            try
            {
                switch (args[0].ToLower())
                {
                    case "generate-wsp":
                    {
                        if (args.Length != 9)
                            throw new Exception("Nine parameters expected.");


                        var numberOfSteps = int.Parse(args[1]);
                        var numberOfUsers = int.Parse(args[2]);
                        var authorisationListsDensity = double.Parse(args[3]);
                        var notEqualsConstraints = int.Parse(args[4]);
                        var atMostConstraints = int.Parse(args[5]);
                        var atLeastConstraints = int.Parse(args[6]);
                        var seed = int.Parse(args[7]);

                        Console.WriteLine(
                            "Generating WSP instance with {0} steps, {1} users, authorisations density {2}, {3} not-equals constraints, {4} at-most-3 out of 5 constraints, {5} at-least-3 out of five constraints, and random number generator seed value {6}...",
                            numberOfSteps, numberOfUsers, authorisationListsDensity, notEqualsConstraints,
                            atMostConstraints, atLeastConstraints, seed);

                        var instance = WspInstance.Random(
                            numberOfSteps: numberOfSteps,
                            numberOfUsers: numberOfUsers,
                            authorisationListsDensity: authorisationListsDensity,
                            notEqualsConstraints: notEqualsConstraints,
                            atMostConstraints: atMostConstraints,
                            atLeastConstraints: atLeastConstraints,
                            seed: seed);

                        Console.WriteLine("WSP instance successfully generated.");


                        var outputFilename = args[8];
                        instance.SaveToFile(outputFilename);
                        Console.WriteLine("WSP instance saved to '{0}'.", outputFilename);
                        break;
                    }

                    case "generate-pbpb":
                    {
                        if (args.Length != 3)
                            throw new Exception("Three parameters expected.");

                        var instanceFilename = args[1];
                        var outputFilename = args[2];

                        var instance = new WspInstance(instanceFilename);

                        Console.WriteLine("WSP instance '{0}' is successfully loaded.", instanceFilename);

                        Dictionary<int, Tuple<int, int>> mDirectory;
                        new PbPbSolver().WriteObp(instance, outputFilename, out mDirectory);
                        Console.WriteLine();
                        Console.WriteLine("PBPB formulation of '{0}' is saved to '{1}'.", instanceFilename,
                            outputFilename);

                        break;
                    }

                    case "generate-udpb":
                    {
                        if (args.Length != 3)
                            throw new Exception("Three parameters expected.");

                        var instanceFilename = args[1];
                        var outputFilename = args[2];

                        var instance = new WspInstance(instanceFilename);

                        Console.WriteLine("WSP instance '{0}' is successfully loaded.", instanceFilename);

                        Dictionary<int, Tuple<int, int>> mDirectory;
                        new UdPbSolver().WriteObp(instance, outputFilename);
                        Console.WriteLine();
                        Console.WriteLine("UDPB formulation of '{0}' is saved to '{1}'.", instanceFilename,
                            outputFilename);

                        break;
                    }

                    case "run-pbt":
                    {
                        if (args.Length != 4)
                            throw new Exception("Four parameters expected.");


                        var instanceFilename = args[1];
                        double timeLimitDouble = double.Parse(args[2]);
                        var solutionFilename = args[3];

                        var timeLimit = timeLimitDouble < 0 ? TimeInterval.Infinity : TimeInterval.Second * timeLimitDouble;

                        var instance = new WspInstance(instanceFilename);

                        Console.WriteLine(
                            "WSP instance '{0}' is successfully loaded.  Solving with PBT, time limit {1}...",
                            instanceFilename, timeLimit);

                        var solution =
                            new Pbt(timeLimit).Solve(
                                instance);
                        solution.SaveToFile(solutionFilename);
                        Console.WriteLine("Solution time: " + solution.SolutionTime);
                        Console.WriteLine("Solution outcome: " + solution.SolutionType);
                        Console.WriteLine("Solution saved to '{0}'.", solutionFilename);
                        break;
                    }

                    default:
                        throw new Exception("Unrecognised parameters.");
                }
            }
            catch (Exception ex)
            {
                Console.ForegroundColor = ConsoleColor.Red;
                Console.WriteLine(ex.Message);
                Console.ForegroundColor = prevColour;

                Console.WriteLine();
                HelpMessage();
            }
        }

        private static void HelpMessage()
	    {
	        Console.WriteLine("Usage:");
	        Console.WriteLine();
	        Console.WriteLine("generate-wsp <num. of steps> <num. of users> <authorisation lists density> <num. of not-equals constraints> <num. of at-most-3 constraints with scope 5> <num. of at-least-3 constraints with scope 5> <random generator seed value> <WSP instance filename> -- to generate a WSP instance with the specified parameters.  Example: \"generate-wsp 18 180 0.5 27 18 18 0 wsp18.txt\".  For standard values of the parameters, refer to the publication mentioned above.");
	        Console.WriteLine();
            Console.WriteLine("generate-pbpb <WSP instance filename> <PB filename> -- to generate a pattern based PB formulation of the problem.  Example: \"generate-pbpb wsp18.txt pbpb.txt\"");
	        Console.WriteLine();
            Console.WriteLine("generate-udpb <WSP instance filename> <PB filename> -- to generate a user-driven PB formulation of the problem.  Example: \"generate-udpb wsp18.txt udpb.txt\"");
	        Console.WriteLine();
            Console.WriteLine("run-pbt <WSP instance filename> <time limit in seconds> <output filename> -- to run the Pattern Backtracking algorithm with a given time limit.  Example: \"run-pbt wsp18.txt -1 solution.txt\".  Time limit can be fractional, or -1 for infinity.");
	    }
	}
}
