﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

namespace Wsp.Types
{
    public static class Helper
    {
        public static readonly FileInfo ProgramFile = new FileInfo(Assembly.GetExecutingAssembly().Location);
        public static readonly DirectoryInfo ProgramDirectory = ProgramFile.Directory;

        public static string ListToString(this IEnumerable list, string separator)
        {
            IEnumerator enumerator = list.GetEnumerator();
            if (!enumerator.MoveNext())
                return "";

            var builder = new StringBuilder(enumerator.Current.ToString());
            while (enumerator.MoveNext())
            {
                builder.Append(separator);
                builder.Append(enumerator.Current);
            }

            return builder.ToString();
        }

        public static string ListToString<T>(this IEnumerable<T> list, Func<T, object> selector, string separator)
        {
            IEnumerator<T> enumerator = list.GetEnumerator();
            if (!enumerator.MoveNext())
                return "";

            var builder = new StringBuilder(selector(enumerator.Current).ToString());
            while (enumerator.MoveNext())
            {
                builder.Append(separator);
                builder.Append(selector(enumerator.Current));
            }

            return builder.ToString();
        }


        public static int Random(Random random, int lessThan, int except)
        {
            int result = random.Next(lessThan - 1);

            result = result == except
                ? lessThan - 1
                : result;

            Debug.Assert(result != except);
            Debug.Assert(result >= 0);
            Debug.Assert(result < lessThan);
            return result;
        }

        public static int Except(this Random random, int lessThan, int except)
        {
            Debug.Assert(lessThan > 1);

            int result = random.Next(lessThan - 1);

            result = result == except
                ? lessThan - 1
                : result;

            Debug.Assert(result != except);
            Debug.Assert(result >= 0);
            Debug.Assert(result < lessThan);
            return result;
        }

        public static int Random(Random random, int lessThan, int except1, int except2)
        {
            if (except1 == except2)
                return Random(random, lessThan, except1);

            if (except1 == lessThan - 1)
                return Random(random, lessThan - 1, except2);

            int result1 = Random(random, lessThan - 1, except1);
            int result2 = result1 == except2
                ? lessThan - 1
                : result1;

            Debug.Assert(result2 != except1);
            Debug.Assert(result2 != except2);
            Debug.Assert(result2 >= 0);
            Debug.Assert(result2 < lessThan);
            return result2;
        }

        public static int Except(this Random random, int lessThan, int except1, int except2)
        {
            return Random(random, lessThan, except1, except2);
        }


        public static int Sqr(int x)
        {
            return x * x;
        }


        public static DirectoryInfo GetSubdirectory(DirectoryInfo parent, string name)
        {
            DirectoryInfo[] directories = parent.GetDirectories(name);
            switch (directories.Length)
            {
                case 0:
                    return parent.CreateSubdirectory(name);

                case 1:
                    return directories[0];

                default:
                    throw new Exception("Several directories with the same name found.");
            }
        }

        public static string GetFullFilename(DirectoryInfo directory, string filename)
        {
            return directory.FullName + '\\' + filename;
        }

        public static string GetFullFilename(string filename)
        {
            return GetFullFilename(ProgramDirectory, filename);
        }

        public static string GetFullFilename(DirectoryInfo directory, string filename, string extension)
        {
            return GetFullFilename(directory, filename + extension);
        }


        public static int[] DistinctUnordered(this Random random, int lessThan, int numberOfElements)
        {
            Debug.Assert(numberOfElements <= lessThan);
            switch (numberOfElements)
            {
                case 0:
                    return new int[0];

                case 1:
                    return new[] {random.Next(lessThan)};

                case 2:
                {
                    var result = new int[2];
                    result[0] = random.Next(lessThan);
                    result[1] = random.Except(lessThan, result[0]);
                    return result;
                }

                case 3:
                {
                    var result = new int[3];
                    result[0] = random.Next(lessThan);
                    result[1] = random.Except(lessThan, result[0]);
                    result[2] = random.Except(lessThan, result[0], result[1]);
                    return result;
                }

                default:
                {
                    if (2 * numberOfElements >= lessThan)
                    {
                        // O(numberOfElements * lessThan) iterations; O(numberOfElements + lessThan) memory
                        var available = new Range(lessThan).ToList();
                        var result = new int[numberOfElements];
                        for (int i = 0; i < numberOfElements; i++)
                        {
                            int index = random.Next(available.Count);
                            result[i] = available[index];
                            available.RemoveAt(index);
                        }

                        return result;
                    }
                    else
                    {
                        // O(numberOfElements^2) iterations if (numberOfElements / lessThan < const < 1); O(numberOfElements) memory
                        var result = new int[numberOfElements];
                        for (int i = 0; i < numberOfElements; i++)
                        {
                            int value;
                            do
                            {
                                value = random.Next(lessThan);
                            } while (Array.IndexOf(result, value, 0, i) >= 0);

                            result[i] = value;
                        }

                        return result.ToArray();
                    }
                }
            }
        }


        public static T[] FillArray<T>(this T fillWith, int length)
        {
            var result = new T[length];
            for (int i = result.Length - 1; i >= 0; i--)
                result[i] = fillWith;

            return result;
        }

        public static void FillExistingArray<T>(T[] array, T fillWith)
        {
            for (int i = array.Length - 1; i >= 0; i--)
                array[i] = fillWith;
        }
    }
}
