﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.Linq;

namespace Wsp.Types
{
	public struct Int64BasedSet : IEnumerable<int>, IEquatable<Int64BasedSet>
	{
		private long bits;

		public Int64BasedSet(params IConvertible[] values)
		{
			bits = 0;
			for (int i = values.Length - 1; i >= 0; i--)
				this[(int)values[i]] = true;
		}

		public Int64BasedSet(IEnumerable<int> values)
		{
			bits = 0;
			foreach (var value in values)
				this[value] = true;
		}

		public Int64BasedSet(params int[] values)
		{
			bits = 0;
			for (int i = values.Length - 1; i >= 0; i--)
				this[values[i]] = true;
		}

		public Int64BasedSet(long bits)
		{
			this.bits = bits;
		}

		public static Int64BasedSet OneElementSet(int element)
		{
			return new Int64BasedSet(((long)1) << element);
		}

		public bool this[int index]
		{
			get
			{
				Debug.Assert(index >= 0 && index < 64);
				return (bits & ((long)1 << index)) != 0;
			}
			set
			{
				Debug.Assert(index >= 0 && index < 64);
				if (value)
					bits |= (long)1 << index;
				else
					bits &= ~((long)1 << index);
			}
		}

        [Pure]
		public Int64BasedSet Intersection(Int64BasedSet other)
		{
			return new Int64BasedSet(bits & other.bits);
		}

        [Pure]
        public bool Intersects(Int64BasedSet other)
		{
			return (bits & other.bits) != 0;
		}

        [Pure]
        public Int64BasedSet Union(Int64BasedSet other)
		{
			return new Int64BasedSet(bits | other.bits);
		}

		public int NumberOfElements
		{
			get
			{
				int count;
				long value = bits;
				for (count = 0; value != 0; count++)
					value &= value - 1;

				var temp = bits - ((bits >> 1) & 0x5555555555555555);
				temp = (temp & 0x3333333333333333) + ((temp >> 2) & 0x3333333333333333);
				var result = (int)((((temp + (temp >> 4)) & 0xF0F0F0F0F0F0F0F) * 0x101010101010101) >> 56);
				
				if (result != count)
					throw new Exception("Bit counting is wrong.");

				return result;
			}
		}


		public long Bits
		{
			get { return bits; }
		}

		public bool IsEmpty
		{
			get { return bits == 0; }
		}

		public bool ContainsAtMostOneElement
		{
			get { return (bits & (bits - 1)) == 0; }
		}

		public static IEnumerable<Int64BasedSet> AllSubsets(int n)
		{
			for (int i = 0; i < (1 << n); i++)
				yield return new Int64BasedSet(i);
		}

		public IEnumerator<int> GetEnumerator()
		{
			for (int i = 0; i < 64; i++)
				if (this[i])
					yield return i;
		}

		public override string ToString()
		{
			Int64BasedSet thisCopy = this;
			return string.Format("{{{0}}}", new Range(0, 63).Where(i => thisCopy[i]).ListToString(", "));
		}

		IEnumerator IEnumerable.GetEnumerator()
		{
			return GetEnumerator();
		}

		public static Int64BasedSet All(int elementsCount)
		{
			Debug.Assert(elementsCount >= 0 && elementsCount <= 64);
            if (elementsCount == 64)
                return new Int64BasedSet((long)-1);

			return new Int64BasedSet(((long)1 << elementsCount) - 1);
		}

/*
		public int Random(Random random)
		{
			if (bits == 0)
				throw new LabException("The int based set is empty.");

			var count = NumberOfElements;
			Debug.Assert(count > 0);

			int index = random.Next(count);
			for (int i = 0; ; i++)
				if (this[i] && index-- == 0)
					return i;
		}
*/

		public static bool operator ==(Int64BasedSet set1, Int64BasedSet set2)
		{
			return set1.Bits == set2.Bits;
		}

		public static bool operator !=(Int64BasedSet set1, Int64BasedSet set2)
		{
			return set1.Bits != set2.Bits;
		}

		public bool Equals(Int64BasedSet other)
		{
			return bits == other.bits;
		}

		public override bool Equals(object obj)
		{
			if (ReferenceEquals(null, obj))
				return false;

			return obj is Int64BasedSet && Equals((Int64BasedSet)obj);
		}

		public override int GetHashCode()
		{
			return (int)(bits ^ (bits >> 32));
		}

		public bool IsSuperset(Int64BasedSet subset)
		{
			return (bits & subset.bits) == subset.bits;
		}

		public Int64BasedSet Exclude(Int64BasedSet toExclude)
		{
			return new Int64BasedSet(bits & (~toExclude.bits));
		}


		public int IndexOfFirstElement()
		{
			Debug.Assert(!IsEmpty);
			
			long b = bits;
			int result = 0;
			while ((b & 1) == 0)
			{
				b >>= 1;
				result++;
			}

			return result;
		}

		public Int64BasedSet ExtractFirstElement()
		{
			Debug.Assert(!IsEmpty);

			return new Int64BasedSet(bits & ~(bits - 1));
		}

		public static readonly Int64BasedSet Empty = new Int64BasedSet();
	}
}
