﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text.RegularExpressions;

namespace Wsp.Types
{
	class WspInstance
	{
		public readonly int NumberOfSteps;
		public readonly int NumberOfUsers;
		public readonly Int64BasedSet[] AuthorisationLists;
		public readonly NotEqualsConstraint[] NotEquals;
		public readonly GenericConstraint[] AtMost;
		public readonly GenericConstraint[] AtLeast;

		public WspInstance(string filename)
		{
			using (var reader = File.OpenText(filename))
			{
				NumberOfSteps = ReadIntAttribute(reader, "#Steps");
				NumberOfUsers = ReadIntAttribute(reader, "#Users");
				int numberOfConstraints = ReadIntAttribute(reader, "#Constraints");

				ReadHeader(reader, "Authorizations:");
				AuthorisationLists = new Int64BasedSet[NumberOfUsers];
				for (int i = 0; i < NumberOfUsers; i++)
				{
					var listStr = ReadStrAttribute(reader, "User " + (i + 1));
					var strs = listStr.Split(new[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
					if (strs.Length != NumberOfSteps)
						throw new Exception("");

					var list = new Int64BasedSet();
					for (int j = 0; j < NumberOfSteps; j++)
					{
						switch (strs[j])
						{
							case "0":
								break;

							case "1":
								list[j] = true;
								break;

							default:
								throw new Exception(string.Format("Cannot parse authorisation list element '{0}'.", strs[j]));
						}
					}

					AuthorisationLists[i] = list;
				}

				var notEqualsList = new List<NotEqualsConstraint>();
				var atMostList = new List<GenericConstraint>();
				var atLeastList = new List<GenericConstraint>();

				ReadHeader(reader, "Constraints:");
				for (int i = 0; i < numberOfConstraints; i++)
				{
					var line = reader.ReadLine();
					if (line == null)
						throw new Exception(string.Format("Constant {0} expected but EOF found.", i));

					var match = Regex.Match(line.ToLower(), @"^(at least|at most)\s+(\d+)\s+users for ([\d\s\,]+)$");
					if (!match.Success)
						throw new Exception(string.Format("Failed to parse constraint '{0}'.", line));

					bool atLeast;
					switch (match.Groups[1].Value)
					{
						case "at least":
							atLeast = true;
							break;

						case "at most":
							atLeast = false;
							break;

						default:
							throw new Exception(string.Format("Unknown constraint type: {0}.", match.Groups[1].Value));
					}

					int numberOfUsers = int.Parse(match.Groups[2].Value);

					var tasks = match.Groups[3].Value.Split(new[] { ' ', ',' }, StringSplitOptions.RemoveEmptyEntries).Select(str => int.Parse(str) - 1).ToArray();

					if (atLeast && tasks.Length == 2 && numberOfUsers == 2)
						notEqualsList.Add(new NotEqualsConstraint(tasks[0], tasks[1]));
					else if (atLeast)
						atLeastList.Add(new GenericConstraint(tasks, numberOfUsers));
					else
						atMostList.Add(new GenericConstraint(tasks, numberOfUsers));
				}

				NotEquals = notEqualsList.ToArray();
				AtLeast = atLeastList.ToArray();
				AtMost = atMostList.ToArray();
			}
		}

		public WspInstance(
			int numberOfSteps,
			int numberOfUsers,
			Int64BasedSet[] authorisationLists,
			NotEqualsConstraint[] notEquals,
			GenericConstraint[] atMost,
			GenericConstraint[] atLeast)
		{
			NumberOfSteps = numberOfSteps;
			NumberOfUsers = numberOfUsers;
			AuthorisationLists = authorisationLists;
			NotEquals = notEquals;
			AtLeast = atLeast;
			AtMost = atMost;
		}

		private static void ReadHeader(StreamReader reader, string title)
		{
			var line = reader.ReadLine();
			if (line == null)
				throw new Exception(string.Format("'{0}' expected but end of file found.", title));

			line = line.Trim().ToLower();
			if (line != title.ToLower())
				throw new Exception(string.Format("'{0}' header expected in line '{1}'.", title, line));
		}

		private static int ReadIntAttribute(StreamReader reader, string name)
		{
			var value = ReadStrAttribute(reader, name);
			int result;
			if (!int.TryParse(value, out result))
				throw new Exception(string.Format("Cannot parse {0} attrbute value '{1}'.", name, value));

			return result;
		}

		private static string ReadStrAttribute(StreamReader reader, string name)
		{
			var line = reader.ReadLine();
			if (line == null)
				throw new Exception(string.Format("'{0}' expected but end of file found.", name));

			line = line.Trim().ToLower();
			if (!line.StartsWith(name.ToLower() + ":"))
				throw new Exception(string.Format("'0' attribute expected in line '{1}'.", name, line));

			return line.Substring(name.Length + 1).Trim();
		}

		public static WspInstance Random(
			int numberOfSteps,
			int numberOfUsers,
			double authorisationListsDensity,
			int notEqualsConstraints,
			int atMostConstraints,
			int atLeastConstraints,
			int seed)
		{
			var maxAuthorisationsNumber = (int)(numberOfSteps * authorisationListsDensity + 0.5);

			var random = new Random(numberOfSteps + numberOfUsers + maxAuthorisationsNumber + notEqualsConstraints + atMostConstraints + atLeastConstraints + seed);
			
			var authorisationLists = new Int64BasedSet[numberOfUsers];
			for (int user = 0; user < numberOfUsers; user++)
			{
				var numberOfAuthorisations = random.Next(1, maxAuthorisationsNumber);
				foreach (var step in random.DistinctUnordered(numberOfSteps, numberOfAuthorisations))
					authorisationLists[user][step] = true;
			}

			if (notEqualsConstraints < 0 || notEqualsConstraints > numberOfSteps * (numberOfSteps - 1) / 2)
				throw new Exception("Incorrect number of not-equals constraints.");

			var notEquals = new NotEqualsConstraint[notEqualsConstraints];
			var notEqualsAssigned = new bool[numberOfSteps, numberOfSteps];
			for (int i = 0; i < notEquals.Length; i++)
			{
				notEquals[i] = new NotEqualsConstraint();
				int step1, step2;
				do
				{
					step1 = random.Next(numberOfSteps);
					step2 = Helper.Random(random, numberOfSteps, step1);
				} while (notEqualsAssigned[step1, step2]);

				notEquals[i] = new NotEqualsConstraint(step1, step2);
				notEqualsAssigned[step1, step2] = true;
				notEqualsAssigned[step2, step1] = true;
			}

			var atLeast = ProduceGenericConstraints(atLeastConstraints, numberOfSteps, random);
			var atMost = ProduceGenericConstraints(atMostConstraints, numberOfSteps, random);

			return new WspInstance(
				numberOfSteps,
				numberOfUsers,
				authorisationLists,
				notEquals,
				atMost,
				atLeast);
		}

		private static GenericConstraint[] ProduceGenericConstraints(int count, int numberOfSteps, Random random)
		{
			var constraints = new List<GenericConstraint>();
			for (int i = 0; i < count; i++)
			{
				GenericConstraint constraint;
				do
				{
					constraint = new GenericConstraint(random.DistinctUnordered(numberOfSteps, 5).ToArray(), 3);
				} while (constraints.Contains(constraint));

				constraints.Add(constraint);
			}

			return constraints.ToArray();
		}

		public void SaveToFile(DirectoryInfo directory)
		{
			SaveToFile(Helper.GetFullFilename(directory, ToString(), ".wsp"));
		}

		public void SaveToFile(string filename)
		{
			using (var writer = File.CreateText(filename))
			{
				writer.WriteLine("#Steps: {0}", NumberOfSteps);
				writer.WriteLine("#Users: {0}", NumberOfUsers);
				writer.WriteLine("#Constraints: {0}", NotEquals.Length + AtMost.Length + AtLeast.Length);
				writer.WriteLine("Authorizations:");
				for (int user = 0; user < AuthorisationLists.Length; user++)
				{
					writer.Write("user {0}:", user + 1);
					for (int step = 0; step < NumberOfSteps; step++)
						writer.Write(AuthorisationLists[user][step] ? " 1" : " 0");

					writer.WriteLine();
				}

				writer.WriteLine("Constraints:");
				foreach (var notEqualsConstraint in NotEquals)
					writer.WriteLine("at least 2 users for {0}, {1}", notEqualsConstraint.Step1 + 1, notEqualsConstraint.Step2 + 1);

				foreach (var atMostConstraint in AtMost)
					writer.WriteLine("at most {0} users for {1}", atMostConstraint.Users, atMostConstraint.Steps.Select(t => t + 1).ListToString(", "));

				foreach (var atLeastConstraint in AtLeast)
					writer.WriteLine("at least {0} users for {1}", atLeastConstraint.Users, atLeastConstraint.Steps.Select(t => t + 1).ListToString(", "));
			}
		}

		public bool[,] GetNotEqualsMatrix()
		{
			var notEquals = new bool[NumberOfSteps, NumberOfSteps];
			for (int i = 0; i < NotEquals.Length; i++)
			{
				notEquals[NotEquals[i].Step1, NotEquals[i].Step2] = true;
				notEquals[NotEquals[i].Step2, NotEquals[i].Step1] = true;
			}
			return notEquals;
		}
	}
}
