﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;

namespace Wsp.Types
{
	class WspSolution
	{
	    private readonly WspInstance instance;
	    public readonly int[] AssignedUsers;
		public WspSolutionType SolutionType { get; private set; }
		public TimeInterval SolutionTime { get; set; }

		public WspSolution(WspInstance instance, TimeInterval solutionTime, WspSolutionType solutionType = WspSolutionType.Unknown)
		{
			SolutionTime = solutionTime;
			SolutionType = solutionType;
			AssignedUsers = new int[instance.NumberOfSteps];
		}

		public WspSolution(WspInstance instance, TimeInterval solutionTime, int[] assignment)
		{
			SolutionTime = solutionTime;
		    this.instance = instance;
		    AssignedUsers = assignment;
			SolutionType = WspSolutionType.Sat;
		}


	    public int DistinctUsers(int[] tasks)
		{
			var involved = new List<int>();

			// Count the number of distinct users (the usedArray is assumed to be reset)
			for (int i = 0; i < tasks.Length; i++)
			{
				int user = AssignedUsers[tasks[i]];
				if (!involved.Contains(user))
					involved.Add(user);
			}

			return involved.Count;
		}

		public bool IsFeasible()
		{
			for (int i = AssignedUsers.Length - 1; i >= 0; i--)
			{
				var user = AssignedUsers[i];
				if (user < 0 || user >= instance.NumberOfUsers)
				{
					Console.WriteLine("User index is outside of the [0, {0}] range.", instance.NumberOfUsers - 1);
					return false;
				}

				if (!instance.AuthorisationLists[user][i])
				{
                    Console.WriteLine("The user in unauthorised.");
					return false;
				}
			}

			for (int i = instance.NotEquals.Length - 1; i >= 0; i--)
			{
				int user1 = AssignedUsers[instance.NotEquals[i].Step1];
				int user2 = AssignedUsers[instance.NotEquals[i].Step2];
				if (user1 == user2)
				{
                    Console.WriteLine("Not-equals violation.");
					return false;
				}
			}

			for (int i = instance.AtMost.Length - 1; i >= 0; i--)
			{
				int count = DistinctUsers(instance.AtMost[i].Steps);
				if (count > instance.AtMost[i].Users)
				{
                    Console.WriteLine("At-most violation.");
					return false;
				}
			}

			for (int i = instance.AtLeast.Length - 1; i >= 0; i--)
			{
				int count = DistinctUsers(instance.AtLeast[i].Steps);
				if (count < instance.AtLeast[i].Users)
				{
                    Console.WriteLine("At-least violation.");
					return false;
				}
			}

			return true;
		}

		public static WspSolution RandomSolution(WspInstance instance, Random random)
		{
			var solution = new WspSolution(instance, TimeInterval.Zero, WspSolutionType.Unknown);
			for (int i = 0; i < instance.NumberOfSteps; i++)
				solution.AssignedUsers[i] = random.Next(instance.NumberOfUsers);

			return solution;
		}

		public static WspSolution ReadFromFile(WspInstance instance, string filename)
		{
			using (var reader = File.OpenText(filename))
			{
				var firstLine = reader.ReadLine();
				if (firstLine == null)
					throw new Exception(string.Format("Empty solution file: {0}.", filename));

				var secondLine = reader.ReadLine();
				if (secondLine == null)
					throw new Exception(string.Format("No solution time provided in solution '{0}'.", filename));

				var solutionTime = TimeInterval.FromSeconds(double.Parse(secondLine));

				switch (firstLine.Trim().ToLower())
				{
					case "unsat":
						return new WspSolution(instance, solutionTime, WspSolutionType.ProvenUnsat);

					case "unknown":
						return new WspSolution(instance, solutionTime, WspSolutionType.Unknown);

					case "sat":
						break;

					default:
						throw new Exception(string.Format("Unknown solution type in file {0}: '{1}'.", filename, firstLine));
				}

				var solution = new WspSolution(instance, solutionTime, WspSolutionType.Sat);
				for (int step = 0; step < instance.NumberOfSteps; step++)
				{
					var line = reader.ReadLine();
					if (line == null)
						throw new Exception(string.Format("Step {0} expected but EOF found in {1}.", step + 1, filename));

					var match = Regex.Match(line, @"^Step (\d+) \-\> User (\d+)$");
					if (!match.Success)
						throw new Exception(string.Format("Incorrect format in file {0}, step '{1}'.", filename, step + 1));

					if (int.Parse(match.Groups[1].Value) != step + 1)
						throw new Exception(string.Format("Step {0} expected (solution {1}).", step + 1, filename));

					var user = int.Parse(match.Groups[2].Value) - 1;
					if (user < 0 || user >= instance.NumberOfUsers)
						throw new Exception(string.Format("Incorrect user specified for step {0} in solution {1}.", step + 1, filename));

					solution.AssignedUsers[step] = user;
				}

				if (!solution.IsFeasible())
					throw new Exception(string.Format("Infeasible solution: {0}.", filename));

				return solution;
			}
		}

		public void SaveToFile(string filename)
		{
			if (SolutionType == WspSolutionType.Sat && !IsFeasible())
				throw new Exception(string.Format("Infeasible solution: {0}.", filename));

			using (var writer = File.CreateText(filename))
			{
				switch (SolutionType)
				{
					case WspSolutionType.Unknown:
						writer.WriteLine("unknown");
						break;

					case WspSolutionType.Sat:
						writer.WriteLine("sat");
						break;

					case WspSolutionType.ProvenUnsat:
						writer.WriteLine("unsat");
						break;

					default:
						throw new Exception(string.Format("Unknown solution type: {0}.", SolutionType));
				}

				writer.WriteLine(SolutionTime.Seconds);

				if (SolutionType == WspSolutionType.Sat)
					for (int i = 0; i < AssignedUsers.Length; i++)
						writer.WriteLine("Step {0} -> User {1}", i + 1, AssignedUsers[i] + 1);
			}
		}
	}
}
