from __future__ import print_function
import collections
from ortools.sat.python import cp_model

def parseAttribute(f, attribute):
    line = f.readline().split(':')
    if line[0] != attribute:
        raise Exception('Attribute mismatch: "' + attribute + '" expected but "' + line[0] + '" found')

    return line[1]

def readHeader(f, header):
    line = f.readline().strip()
    if line != header:
        raise Exception('Header mismatch: "' + header + '" expected but "' + line + '" found')

    return line[1]

def main():
    import sys
    import time
    
    start = time.time()

    f = open(sys.argv[1], 'r')
    
    numOfSteps = int(parseAttribute(f, '#Steps'))
    numOfUsers = int(parseAttribute(f, '#Users'))
    numOfConstraints = int(parseAttribute(f, '#Constraints'))

    readHeader(f, 'Authorizations:')
    authorisations = []
    for i in range(1, numOfUsers + 1):
        authorisations.append(list(int(v) for v in parseAttribute(f, 'user ' + str(i)).strip()))

    #for i in range(numOfUsers): 
    #    print('User' + str(i))
    #    for j in range(numOfSteps):
    #        print(authorisations[i][j])
    
    model = cp_model.CpModel()
    # Create the variables.
    steps = []
    for step in range(numOfSteps):
        bounds = []
        for user in range(numOfUsers):
            if step in range(numOfSteps):
                if authorisations[user][step] == 0:
                    continue;
                
                if len(bounds) > 0 and bounds[len(bounds) - 1] == user - 1:
                    bounds[len(bounds) - 1] = user
                else:
                    bounds.append(user)
                    bounds.append(user)

        steps.append(model.NewEnumeratedIntVar(bounds, 'step' + str(step)))

    readHeader(f, 'Constraints:')

    varCounter = 0

    import re
    for i in range(0, numOfConstraints):
        line = f.readline();
        m = re.match(r'at (least|most) (\d+) users for (.+)', line)

        if m.group(1) == 'least':
            atLeast = True
        elif m.group(1) == 'most':
            atLeast = False
        else:
            raise Exception('Unknown constraint type: ' + main.group(1))

        u = int(m.group(2))
        scope = [int(v) - 1 for v in m.group(3).split(',')]
        #print(scope)

        if atLeast and u == 2 and len(scope) == 2:
            # not-equals
            model.Add(steps[scope[0]] != steps[scope[1]])

        elif atLeast and u == 3 and len(scope) == 5:
            # at-least-3 out of 5
            vars = []
            for i in range(0, 5):
                for j in range(i + 1, 5):
                    for k in range(j + 1, 5):
                        v = model.NewBoolVar('var' + str(varCounter))
                        varCounter += 1
                        # model.Add(((steps[scope[i]] != steps[scope[j]]) and (steps[i] != steps[k]) and (steps[scope[j]] == steps[scope[k]])).OnlyEnforceIf(v))
                        
                        model.Add(steps[scope[i]] != steps[scope[j]]).OnlyEnforceIf(v)
                        model.Add(steps[scope[i]] != steps[scope[k]]).OnlyEnforceIf(v)
                        model.Add(steps[scope[j]] != steps[scope[k]]).OnlyEnforceIf(v)
                        
                        # model.AddAllDifferent([steps[scope[i]], steps[scope[j]], steps[scope[k]]]).OnlyEnforceIf(v)
                        vars.append(v)                    
                        # print(str(i) + ' ' + str(j) + ' ' + str(k))
            
            model.AddBoolOr(vars);

        elif not atLeast and u == 3 and len(scope) == 5:
            # at-most-3 out of 5
            for i in range(5): # We'll skip the i'th step in the scope
                listOfVars = []

                for j in range(5):
                    if j == i:
                        continue;

                    for k in range(j + 1, 5):
                        if k == i:
                            continue;

                        v = model.NewBoolVar('var' + str(varCounter))
                        varCounter += 1

                        model.Add(steps[scope[j]] == steps[scope[k]]).OnlyEnforceIf(v);
                        listOfVars.append(v)

                model.AddBoolOr(listOfVars);


        else:
            raise Exception('Unsupported constraint: {} {} out of {}'.format('at-least' if atLeast else 'at-most', u, len(scope)))

    f.close()

    solver = cp_model.CpSolver()
    status = solver.Solve(model)

    solutionTime = time.time() - start

    f = open(sys.argv[2], "w")
    if status == 2:
        f.write('sat\n')
        f.write(str(solutionTime) + '\n')
        for i in range(numOfSteps):
            f.write('Step {} -> User {}\n'.format(i + 1, solver.Value(steps[i]) + 1))
    elif status == 3:
        f.write('unsat\n')
        f.write(str(solutionTime) + '\n')
    else:
        print('\nUnknown status: {} ({})'.format(solver.StatusName(status), status))
    f.close()


if __name__ == '__main__':
    main()
